@@ -18,8 +18,7 @@ import subprocess
 # extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
 # ones.
 extensions = [
-    'sphinx.ext.autodoc',
-    'sphinx.ext.coverage',
+    'sphinx.ext.extlinks',
     'sphinx.ext.ifconfig',
     'sphinx_rtd_theme',
 ]
@@ -88,3 +87,15 @@ latex_documents = [
   ('index', 'rauc.tex', u'RAUC Documentation',
    author, 'manual'),
 ]
+
+# -- Options for manual page output ---------------------------------------
+
+man_pages = [
+  ('man', 'rauc', 'safe and secure updating', '', 1),
+]
+
+# -- Options for external links -------------------------------------------
+
+extlinks = {
+  'pr': ('https://github.com/rauc/rauc/pull/%s', '#%s'),
+}
