@@ -2,30 +2,52 @@
 
 #include <glib.h>
 
-typedef struct _RaucDMVerity {
+typedef enum _RaucDMType {
+	RAUC_DM_VERITY,
+	RAUC_DM_CRYPT,
+} RaucDMType;
+
+typedef struct _RaucDM {
+	RaucDMType type;
+
+	/* common variables */
 	gchar *uuid;
 	gchar *lower_dev;
 	gchar *upper_dev;
 	guint64 data_size;
+
+	/* dm-verity variables */
 	gchar *root_digest;
 	gchar *salt;
-} RaucDMVerity;
+
+	/* dm-crypt variables */
+	gchar *key;
+} RaucDM;
 
 /**
  * Allocates a new RaucDMVerity with uuid set.
  *
  * @return a pointer to the new RaucDMVerity
  */
-RaucDMVerity *new_dm_verity(void);
+RaucDM *r_dm_new_verity(void);
+
+/**
+ * Allocates a new RaucDMCrypt with uuid set.
+ *
+ * Free with r_dm_free_crypt.
+ *
+ * @return a pointer to the new RaucDMCrypt
+ */
+RaucDM *r_dm_new_crypt(void);
 
 /**
  * Frees the memory allocated by a RaucDMVerity.
  *
  * @param dm_verity struct to free
  */
-void free_dm_verity(RaucDMVerity *dm_verity);
+void r_dm_free(RaucDM *dm_verity);
 
-G_DEFINE_AUTOPTR_CLEANUP_FUNC(RaucDMVerity, free_dm_verity);
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(RaucDM, r_dm_free);
 
 /**
  * Configure a dm-verity target in the kernel using the provided parameters and
@@ -36,7 +58,7 @@ G_DEFINE_AUTOPTR_CLEANUP_FUNC(RaucDMVerity, free_dm_verity);
  *
  * @return TRUE on success, FALSE if an error occurred
  */
-gboolean setup_dm_verity(RaucDMVerity *dm_verity, GError **error);
+gboolean r_dm_setup(RaucDM *dm_verity, GError **error);
 
 /**
  * Remove a previously configured dm-verity target from the kernel.
@@ -47,4 +69,4 @@ gboolean setup_dm_verity(RaucDMVerity *dm_verity, GError **error);
  *
  * @return TRUE on success, FALSE if an error occurred
  */
-gboolean remove_dm_verity(RaucDMVerity *dm_verity, gboolean deferred, GError **error);
+gboolean r_dm_remove(RaucDM *dm_verity, gboolean deferred, GError **error);
