@@ -20,12 +20,14 @@ typedef struct {
 	gchar *status;
 	RaucChecksum checksum;
 	gchar *installed_txn;
-	gchar *installed_timestamp;
+	GDateTime *installed_timestamp;
 	guint32 installed_count;
-	gchar *activated_timestamp;
+	GDateTime *activated_timestamp;
 	guint32 activated_count;
 } RaucSlotStatus;
 
+#define RAUC_FORMAT_ISO_8601 "%Y-%m-%dT%H:%M:%SZ"
+
 typedef struct _RaucSlot {
 	/** name of the slot. A glib intern string. */
 	const gchar *name;
@@ -55,6 +57,8 @@ typedef struct _RaucSlot {
 	guint64 region_start;
 	/** size of both partitions(for boot-mbr-switch, boot-gpt-switch and boot-raw-fallback) */
 	guint64 region_size;
+	/** limit the writable size of the boot partition (for boot-emmc) */
+	guint64 size_limit;
 
 	/** current state of the slot (runtime) */
 	SlotState state;
