@@ -9,6 +9,7 @@
 #include <sys/vfs.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 
 #include <openssl/rand.h>
 #include <unistd.h>
@@ -272,7 +273,7 @@ static gboolean casync_make_arch(const gchar *idxpath, const gchar *contentpath,
 	}
 	g_ptr_array_add(iargs, NULL);
 
-	/* Outer process calll */
+	/* Outer process call */
 	if (!check_pseudo_active()) {
 		g_ptr_array_add(args, g_strdup("fakeroot"));
 	}
@@ -510,7 +511,6 @@ static gchar *convert_tar_extract(RaucImage *image, const gchar *dir, const gcha
 static gchar *convert_composefs(RaucImage *image, const gchar *dir, const gchar *tar_extracted_path, const gchar *fakeroot, GError **error)
 {
 	GError *ierror = NULL;
-	g_autoptr(GPtrArray) args = g_ptr_array_new_full(10, g_free);
 
 	g_return_val_if_fail(image, NULL);
 	g_return_val_if_fail(dir, NULL);
@@ -539,6 +539,7 @@ static gchar *convert_composefs(RaucImage *image, const gchar *dir, const gchar
 		return NULL;
 	}
 
+	g_autoptr(GPtrArray) args = g_ptr_array_new_full(10, g_free);
 	r_fakeroot_add_args(args, fakeroot);
 
 	g_ptr_array_add(args, g_strdup("mkcomposefs"));
@@ -633,7 +634,7 @@ static gboolean convert_images(RaucManifest *manifest, const gchar *dir, const g
 
 		if (tar_extracted_path && !g_strv_contains((const gchar * const *)image->convert, "tar-extract")) {
 			if (!rm_tree(tar_extracted_path, &ierror)) {
-				g_propagate_prefixed_error(error, ierror, "Failed to remove files extacted from tar: ");
+				g_propagate_prefixed_error(error, ierror, "Failed to remove files extracted from tar: ");
 				return FALSE;
 			}
 		}
@@ -709,12 +710,11 @@ static gboolean input_stream_read_bytes_all(GInputStream *stream,
 		GCancellable *cancellable,
 		GError **error)
 {
-	g_autofree void *buffer = NULL;
 	gsize bytes_read;
 
 	g_assert_cmpint(count, !=, 0);
 
-	buffer = g_malloc0(count);
+	g_autofree void *buffer = g_malloc0(count);
 
 	if (!g_input_stream_read_all(stream, buffer, count, &bytes_read,
 			cancellable, error)) {
@@ -728,23 +728,14 @@ static gboolean input_stream_read_bytes_all(GInputStream *stream,
 
 static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, GError **error)
 {
-	g_autoptr(GFile) bundlefile = NULL;
-	g_autoptr(GFileIOStream) bundlestream = NULL;
-	GOutputStream *bundleoutstream = NULL; /* owned by the bundle stream */
 	GError *ierror = NULL;
-	guint64 offset;
-	int bundlefd = -1;
-	guint8 salt[32] = {0};
-	guint8 hash[32] = {0};
-	uint64_t combined_size = 0;
-	guint64 verity_size = 0;
 
 	g_return_val_if_fail(bundlename != NULL, FALSE);
 	g_return_val_if_fail(manifest != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	bundlefile = g_file_new_for_path(bundlename);
-	bundlestream = g_file_open_readwrite(bundlefile, NULL, &ierror);
+	g_autoptr(GFile) bundlefile = g_file_new_for_path(bundlename);
+	g_autoptr(GFileIOStream) bundlestream = g_file_open_readwrite(bundlefile, NULL, &ierror);
 	if (bundlestream == NULL) {
 		g_propagate_prefixed_error(
 				error,
@@ -753,8 +744,9 @@ static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, G
 		return FALSE;
 	}
 
+	GOutputStream *bundleoutstream = NULL; /* owned by the bundle stream */
 	bundleoutstream = g_io_stream_get_output_stream(G_IO_STREAM(bundlestream));
-	bundlefd = g_file_descriptor_based_get_fd(G_FILE_DESCRIPTOR_BASED(bundleoutstream));
+	int bundlefd = g_file_descriptor_based_get_fd(G_FILE_DESCRIPTOR_BASED(bundleoutstream));
 
 	/* check we have a clean manifest */
 	g_assert(manifest->bundle_verity_salt == NULL);
@@ -770,9 +762,10 @@ static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, G
 		return FALSE;
 	}
 
-	offset = g_seekable_tell(G_SEEKABLE(bundlestream));
+	guint64 offset = g_seekable_tell(G_SEEKABLE(bundlestream));
 	g_debug("Payload size: %" G_GUINT64_FORMAT " bytes.", offset);
 	/* dm-verity hash table generation */
+	guint8 salt[32] = {0};
 	if (RAND_bytes((unsigned char *)&salt, sizeof(salt)) != 1) {
 		g_set_error(error,
 				R_BUNDLE_ERROR,
@@ -794,6 +787,9 @@ static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, G
 				"squashfs size (%"G_GUINT64_FORMAT ") must be larger than 4096 bytes", offset);
 		return FALSE;
 	}
+
+	guint8 hash[32] = {0};
+	uint64_t combined_size = 0;
 	if (r_verity_hash_create(bundlefd, offset/4096, &combined_size, hash, salt) != 0) {
 		g_set_error(error,
 				R_BUNDLE_ERROR,
@@ -803,7 +799,7 @@ static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, G
 	}
 	/* for a squashfs <= 4096 bytes, we don't have a hash table */
 	g_assert(combined_size*4096 > (uint64_t)offset);
-	verity_size = combined_size*4096 - offset;
+	guint64 verity_size = combined_size*4096 - offset;
 	g_assert(verity_size % 4096 == 0);
 
 	manifest->bundle_verity_salt = r_hex_encode(salt, sizeof(salt));
@@ -815,18 +811,14 @@ static gboolean create_verity(const gchar *bundlename, RaucManifest *manifest, G
 
 static gboolean append_signature_to_bundle(const gchar *bundlename, GBytes *sig, GError **error)
 {
-	g_autoptr(GFile) bundlefile = NULL;
-	g_autoptr(GFileIOStream) bundlestream = NULL;
-	GOutputStream *bundleoutstream = NULL; /* owned by the bundle stream */
 	GError *ierror = NULL;
-	guint64 offset;
 
 	g_return_val_if_fail(bundlename, FALSE);
 	g_return_val_if_fail(sig, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	bundlefile = g_file_new_for_path(bundlename);
-	bundlestream = g_file_open_readwrite(bundlefile, NULL, &ierror);
+	g_autoptr(GFile) bundlefile = g_file_new_for_path(bundlename);
+	g_autoptr(GFileIOStream) bundlestream = g_file_open_readwrite(bundlefile, NULL, &ierror);
 	if (bundlestream == NULL) {
 		g_propagate_prefixed_error(
 				error,
@@ -835,6 +827,7 @@ static gboolean append_signature_to_bundle(const gchar *bundlename, GBytes *sig,
 		return FALSE;
 	}
 
+	GOutputStream *bundleoutstream = NULL; /* owned by the bundle stream */
 	bundleoutstream = g_io_stream_get_output_stream(G_IO_STREAM(bundlestream));
 
 	if (!g_seekable_seek(G_SEEKABLE(bundlestream),
@@ -846,7 +839,7 @@ static gboolean append_signature_to_bundle(const gchar *bundlename, GBytes *sig,
 		return FALSE;
 	}
 
-	offset = g_seekable_tell(G_SEEKABLE(bundlestream));
+	guint64 offset = g_seekable_tell(G_SEEKABLE(bundlestream));
 	g_debug("Signature offset: %" G_GUINT64_FORMAT " bytes.", offset);
 	if (!output_stream_write_bytes_all(bundleoutstream, sig, NULL, &ierror)) {
 		g_propagate_prefixed_error(
@@ -876,18 +869,41 @@ static gboolean append_signature_to_bundle(const gchar *bundlename, GBytes *sig,
 	return TRUE;
 }
 
-static GBytes *generate_bundle_signature(const gchar *bundlename, RaucManifest *manifest, GError **error)
+static GBytes *append_signature(GBytes *input_sig, GError **error)
 {
 	GError *ierror = NULL;
-	g_autoptr(GBytes) sig = NULL;
 
-	g_return_val_if_fail(bundlename, FALSE);
-	g_return_val_if_fail(manifest, FALSE);
+	g_return_val_if_fail(input_sig, NULL);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
 	g_assert_nonnull(r_context()->certpath);
 	g_assert_nonnull(r_context()->keypath);
 
+	g_autoptr(GBytes) output_sig = cms_append_signature(input_sig,
+			r_context()->certpath,
+			r_context()->keypath,
+			r_context()->intermediatepaths,
+			&ierror);
+	if (output_sig == NULL) {
+		g_propagate_error(error, ierror);
+		return NULL;
+	}
+
+	return g_steal_pointer(&output_sig);
+}
+
+static GBytes *generate_bundle_signature(const gchar *bundlename, RaucManifest *manifest, GError **error)
+{
+	GError *ierror = NULL;
+
+	g_return_val_if_fail(bundlename, NULL);
+	g_return_val_if_fail(manifest, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	g_assert_nonnull(r_context()->certpath);
+	g_assert_nonnull(r_context()->keypath);
+
+	g_autoptr(GBytes) sig = NULL;
 	if (manifest->bundle_format == R_MANIFEST_FORMAT_PLAIN) {
 		if (!check_manifest_internal(manifest, &ierror)) {
 			g_propagate_prefixed_error(
@@ -941,7 +957,6 @@ static GBytes *generate_bundle_signature(const gchar *bundlename, RaucManifest *
 static gboolean sign_bundle(const gchar *bundlename, RaucManifest *manifest, GError **error)
 {
 	GError *ierror = NULL;
-	g_autoptr(GBytes) sig = NULL;
 
 	g_return_val_if_fail(bundlename, FALSE);
 	g_return_val_if_fail(manifest, FALSE);
@@ -957,7 +972,7 @@ static gboolean sign_bundle(const gchar *bundlename, RaucManifest *manifest, GEr
 		}
 	}
 
-	sig = generate_bundle_signature(bundlename, manifest, &ierror);
+	g_autoptr(GBytes) sig = generate_bundle_signature(bundlename, manifest, &ierror);
 	if (!sig) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -983,12 +998,7 @@ static gchar* get_random_file_name(void)
 
 static gboolean encrypt_bundle_payload(const gchar *bundlepath, RaucManifest *manifest, GError **error)
 {
-	gboolean res = FALSE;
-	guint8 key[32] = {0};
 	GError *ierror = NULL;
-	g_autofree gchar* dirname = NULL;
-	g_autofree gchar* tmpfilename = NULL;
-	g_autofree gchar* encpath = NULL;
 
 	g_return_val_if_fail(bundlepath, FALSE);
 	g_return_val_if_fail(manifest, FALSE);
@@ -996,26 +1006,25 @@ static gboolean encrypt_bundle_payload(const gchar *bundlepath, RaucManifest *ma
 
 	g_message("Encrypting bundle payload in aes-cbc-plain64 mode");
 
-	dirname = g_path_get_dirname(bundlepath);
-	tmpfilename = get_random_file_name();
-	encpath = g_build_filename(dirname, tmpfilename, NULL);
+	g_autofree gchar* dirname = g_path_get_dirname(bundlepath);
+	g_autofree gchar* tmpfilename = get_random_file_name();
+	g_auto(RTempFile) encpath = g_build_filename(dirname, tmpfilename, NULL); /* remove on early return */
 
 	/* check we have a clean manifest */
 	g_assert(manifest->bundle_crypt_key == NULL);
 
+	guint8 key[32] = {0};
 	if (RAND_bytes((unsigned char *)&key, sizeof(key)) != 1) {
 		g_set_error(error,
 				R_BUNDLE_ERROR,
 				R_BUNDLE_ERROR_CRYPT,
 				"Failed to generate crypt key");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	res = r_crypt_encrypt(bundlepath, encpath, key, &ierror);
-	if (!res) {
+	if (!r_crypt_encrypt(bundlepath, encpath, key, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	manifest->bundle_crypt_key = r_hex_encode(key, sizeof(key));
@@ -1029,18 +1038,11 @@ static gboolean encrypt_bundle_payload(const gchar *bundlepath, RaucManifest *ma
 				G_FILE_ERROR,
 				g_file_error_from_errno(err),
 				"Renaming %s to %s failed, aborting encryption: %s", encpath, bundlepath, g_strerror(err));
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 	g_clear_pointer(&encpath, g_free); /* prevent removal */
 
-out:
-	/* Remove temporary bundle creation directory */
-	if (encpath)
-		if (g_remove(encpath) != 0)
-			g_warning("Failed to remove temporary encryption file %s", encpath);
-
-	return res;
+	return TRUE;
 }
 
 static gboolean decrypt_bundle_payload(RaucBundle *bundle, GError **error)
@@ -1098,7 +1100,7 @@ out:
 	return res;
 }
 
-static gboolean check_workdir_content(const gchar *path, GError **error)
+static gboolean check_workdir_content(const gchar *path, gboolean top_level, GError **error)
 {
 	GError *ierror = NULL;
 	g_autofree gchar *name = g_path_get_basename(path);
@@ -1133,12 +1135,14 @@ static gboolean check_workdir_content(const gchar *path, GError **error)
 		return TRUE;
 	}
 
-	/* Directories should not be needed in the bundle input, so reject
-	 * them. If you have a use case for this, contact us. */
 	if (g_file_test(path, G_FILE_TEST_IS_DIR)) {
-		g_set_error(error, R_BUNDLE_ERROR, R_BUNDLE_ERROR_PAYLOAD,
-				"directories are not supported as bundle contents (%s)", name);
-		return FALSE;
+		if (top_level && g_str_has_prefix(name, ".")) {
+			g_set_error(error, R_BUNDLE_ERROR, R_BUNDLE_ERROR_PAYLOAD,
+					"hidden directories are not supported as top-level bundle contents (%s)",
+					name);
+			return FALSE;
+		}
+		return TRUE;
 	}
 
 	if (!g_file_test(path, G_FILE_TEST_IS_REGULAR)) {
@@ -1150,12 +1154,67 @@ static gboolean check_workdir_content(const gchar *path, GError **error)
 	return TRUE;
 }
 
+static gboolean link_contentdir_to_workdir(const gchar *contentdir, const gchar *workdir, gboolean top_dir, GError **error)
+{
+	GError *ierror = NULL;
+
+	g_autoptr(GDir) dir = g_dir_open(contentdir, 0, &ierror);
+	if (!dir) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
+	const gchar *name;
+	while ((name = g_dir_read_name(dir))) {
+		if (g_strcmp0(name, ".rauc-workdir") == 0)
+			continue;
+
+		g_autofree gchar *oldpath = g_build_filename(contentdir, name, NULL);
+		if (!check_workdir_content(oldpath, top_dir, &ierror)) {
+			g_propagate_error(error, ierror);
+			return FALSE;
+		}
+
+		g_autofree gchar *newpath = g_build_filename(workdir, name, NULL);
+
+		if (g_file_test(oldpath, G_FILE_TEST_IS_DIR)) {
+			struct stat stat_data = {};
+			if (g_stat(oldpath, &stat_data) < 0) {
+				int err = errno;
+				g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(err),
+						"failed to get file status for '%s': %s", oldpath, g_strerror(err));
+				return FALSE;
+			}
+			if (g_mkdir(newpath, stat_data.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO)) != 0) {
+				int err = errno;
+				g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(err),
+						"failed to create workdir '%s': %s", newpath, g_strerror(err));
+				return FALSE;
+			}
+
+			if (!link_contentdir_to_workdir(oldpath, newpath, FALSE, &ierror)) {
+				g_propagate_error(error, ierror);
+				return FALSE;
+			}
+
+			continue;
+		}
+
+		if (link(oldpath, newpath) != 0) {
+			int err = errno;
+			g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(err),
+					"failed to hard-link file '%s' to workdir: %s", name, g_strerror(err));
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
 static gchar *prepare_workdir(const gchar *contentdir, GError **error)
 {
 	GError *ierror = NULL;
 	g_autofree gchar* workdir = NULL;
-	g_autoptr(GDir) dir = NULL;
-	const gchar *name;
 
 	g_return_val_if_fail(contentdir, NULL);
 	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
@@ -1183,36 +1242,11 @@ static gchar *prepare_workdir(const gchar *contentdir, GError **error)
 		return NULL;
 	}
 
-	dir = g_dir_open(contentdir, 0, &ierror);
-	if (!dir) {
+	if (!link_contentdir_to_workdir(contentdir, workdir, TRUE, &ierror)) {
 		g_propagate_error(error, ierror);
 		return NULL;
 	}
 
-	while ((name = g_dir_read_name(dir))) {
-		g_autofree gchar *oldpath = NULL;
-		g_autofree gchar *newpath = NULL;
-
-		if (g_strcmp0(name, ".rauc-workdir") == 0)
-			continue;
-
-		oldpath = g_build_filename(contentdir, name, NULL);
-
-		if (!check_workdir_content(oldpath, &ierror)) {
-			g_propagate_error(error, ierror);
-			return NULL;
-		}
-
-		newpath = g_build_filename(workdir, name, NULL);
-
-		if (link(oldpath, newpath) != 0) {
-			int err = errno;
-			g_set_error(error, G_FILE_ERROR, g_file_error_from_errno(err),
-					"failed to hard-link file '%s' to workdir: %s", name, g_strerror(err));
-			return NULL;
-		}
-	}
-
 	return g_steal_pointer(&workdir);
 }
 
@@ -1273,6 +1307,8 @@ gboolean create_bundle(const gchar *bundlename, const gchar *contentdir, GError
 		return FALSE;
 	}
 
+	g_auto(RTempFile) bundletmp = g_strdup(bundlename); /* remove on early return */
+
 	workdir = prepare_workdir(contentdir, &ierror);
 	if (!workdir) {
 		g_propagate_error(error, ierror);
@@ -1363,6 +1399,7 @@ gboolean create_bundle(const gchar *bundlename, const gchar *contentdir, GError
 		goto out;
 	}
 
+	g_clear_pointer(&bundletmp, g_free); /* prevent removal */
 	res = TRUE;
 
 out:
@@ -1373,80 +1410,59 @@ out:
 			g_warning("failed to clean up fakeroot environment: %s", cleanup_error->message);
 		}
 	}
-	/* Remove output file on error */
-	if (!res &&
-	    g_file_test(bundlename, G_FILE_TEST_IS_REGULAR))
-		if (g_remove(bundlename) != 0)
-			g_warning("failed to remove %s", bundlename);
 	return res;
 }
 
 static gboolean truncate_bundle(const gchar *inpath, const gchar *outpath, goffset size, GError **error)
 {
-	g_autoptr(GFile) infile = NULL;
-	g_autoptr(GFile) outfile = NULL;
-	g_autoptr(GFileInputStream) instream = NULL;
-	g_autoptr(GFileOutputStream) outstream = NULL;
 	GError *ierror = NULL;
-	gboolean res = FALSE;
-	gssize ssize;
 
 	g_return_val_if_fail(inpath != NULL, FALSE);
 	g_return_val_if_fail(outpath != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	infile = g_file_new_for_path(inpath);
-	outfile = g_file_new_for_path(outpath);
+	g_autoptr(GFile) infile = g_file_new_for_path(inpath);
+	g_autoptr(GFile) outfile = g_file_new_for_path(outpath);
 
-	instream = g_file_read(infile, NULL, &ierror);
+	g_autoptr(GFileInputStream) instream = g_file_read(infile, NULL, &ierror);
 	if (instream == NULL) {
 		g_propagate_prefixed_error(
 				error,
 				ierror,
 				"failed to open bundle for reading: ");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
-	outstream = g_file_create(outfile, G_FILE_CREATE_NONE, NULL,
+	g_autoptr(GFileOutputStream) outstream = g_file_create(outfile, G_FILE_CREATE_NONE, NULL,
 			&ierror);
 	if (outstream == NULL) {
 		g_propagate_prefixed_error(
 				error,
 				ierror,
 				"failed to open bundle for writing: ");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	ssize = g_output_stream_splice(
+	gssize ssize = g_output_stream_splice(
 			(GOutputStream*)outstream,
 			(GInputStream*)instream,
 			G_OUTPUT_STREAM_SPLICE_CLOSE_SOURCE,
 			NULL, &ierror);
 	if (ssize == -1) {
 		g_propagate_error(error, ierror);
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	res = g_seekable_truncate(G_SEEKABLE(outstream), size, NULL, &ierror);
-	if (!res) {
+	if (!g_seekable_truncate(G_SEEKABLE(outstream), size, NULL, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-	res = TRUE;
-out:
-	return res;
+	return TRUE;
 }
 
-gboolean resign_bundle(RaucBundle *bundle, const gchar *outpath, GError **error)
+gboolean resign_bundle(RaucBundle *bundle, const gchar *outpath, gboolean append, GError **error)
 {
-	g_autoptr(RaucManifest) loaded_manifest = NULL;
-	RaucManifest *manifest = NULL; /* alias pointer, not to be freed */
 	GError *ierror = NULL;
-	gboolean res = FALSE;
-	g_autoptr(GBytes) sig = NULL;
 
 	g_return_val_if_fail(bundle != NULL, FALSE);
 	g_return_val_if_fail(outpath != NULL, FALSE);
@@ -1457,51 +1473,50 @@ gboolean resign_bundle(RaucBundle *bundle, const gchar *outpath, GError **error)
 		return FALSE;
 	}
 
-	res = check_bundle_payload(bundle, &ierror);
-	if (!res) {
+	g_auto(RTempFile) outtmp = g_strdup(outpath); /* remove on early return */
+
+	if (!check_bundle_payload(bundle, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
+	g_autoptr(RaucManifest) loaded_manifest = NULL;
+	RaucManifest *manifest = NULL; /* alias pointer, not to be freed */
 	if (bundle->manifest) {
 		manifest = bundle->manifest;
 	} else {
-		res = load_manifest_from_bundle(bundle, &loaded_manifest, &ierror);
-		if (!res) {
+		if (!load_manifest_from_bundle(bundle, &loaded_manifest, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 		manifest = loaded_manifest;
 	}
 
 	g_print("Resigning '%s' format bundle\n", r_manifest_bundle_format_to_str(manifest->bundle_format));
 
-	res = truncate_bundle(bundle->path, outpath, bundle->size, &ierror);
-	if (!res) {
+	if (!truncate_bundle(bundle->path, outpath, bundle->size, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-	sig = generate_bundle_signature(outpath, manifest, &ierror);
+	g_autoptr(GBytes) sig = NULL;
+	if (append) {
+		sig = append_signature(bundle->sigdata, &ierror);
+	} else {
+		sig = generate_bundle_signature(outpath, manifest, &ierror);
+	}
 	if (!sig) {
 		g_propagate_error(error, ierror);
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	res = append_signature_to_bundle(outpath, sig, &ierror);
-	if (!res) {
+	if (!append_signature_to_bundle(outpath, sig, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-out:
-	/* Remove output file on error */
-	if (!res &&
-	    g_file_test(outpath, G_FILE_TEST_IS_REGULAR))
-		if (g_remove(outpath) != 0)
-			g_warning("failed to remove %s", outpath);
-	return res;
+	g_clear_pointer(&outtmp, g_free); /* prevent removal */
+	return TRUE;
 }
 
 static gboolean convert_to_casync_bundle(RaucBundle *bundle, const gchar *outbundle, const gchar **ignore_images, GError **error)
@@ -1646,7 +1661,6 @@ out:
 gboolean create_casync_bundle(RaucBundle *bundle, const gchar *outbundle, const gchar **ignore_images, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	g_return_val_if_fail(bundle != NULL, FALSE);
 	g_return_val_if_fail(outbundle != NULL, FALSE);
@@ -1657,33 +1671,25 @@ gboolean create_casync_bundle(RaucBundle *bundle, const gchar *outbundle, const
 		return FALSE;
 	}
 
-	res = check_bundle_payload(bundle, &ierror);
-	if (!res) {
+	g_auto(RTempFile) outtmp = g_strdup(outbundle); /* remove on early return */
+
+	if (!check_bundle_payload(bundle, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-	res = convert_to_casync_bundle(bundle, outbundle, ignore_images, &ierror);
-	if (!res) {
+	if (!convert_to_casync_bundle(bundle, outbundle, ignore_images, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-	res = TRUE;
-out:
-	/* Remove output file on error */
-	if (!res &&
-	    g_file_test(outbundle, G_FILE_TEST_IS_REGULAR))
-		if (g_remove(outbundle) != 0)
-			g_warning("failed to remove %s", outbundle);
-	return res;
+	g_clear_pointer(&outtmp, g_free); /* prevent removal */
+	return TRUE;
 }
 
 gboolean encrypt_bundle(RaucBundle *bundle, const gchar *outbundle, GError **error)
 {
 	GError *ierror = NULL;
-	g_autoptr(GBytes) encdata = NULL;
-	gboolean res = FALSE;
 
 	g_return_val_if_fail(bundle != NULL, FALSE);
 	g_return_val_if_fail(outbundle != NULL, FALSE);
@@ -1704,39 +1710,30 @@ gboolean encrypt_bundle(RaucBundle *bundle, const gchar *outbundle, GError **err
 		return FALSE;
 	}
 
-	res = truncate_bundle(bundle->path, outbundle, bundle->size, &ierror);
-	if (!res) {
+	g_auto(RTempFile) outtmp = g_strdup(outbundle); /* remove on early return */
+
+	if (!truncate_bundle(bundle->path, outbundle, bundle->size, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* encrypt sigdata CMS */
-	encdata = cms_encrypt(bundle->sigdata, r_context()->recipients, &ierror);
+	g_autoptr(GBytes) encdata = cms_encrypt(bundle->sigdata, r_context()->recipients, &ierror);
 	if (encdata == NULL) {
 		g_propagate_prefixed_error(
 				error,
 				ierror,
 				"Failed to encrypt bundle: ");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	res = append_signature_to_bundle(outbundle, encdata, &ierror);
-	if (!res) {
+	if (!append_signature_to_bundle(outbundle, encdata, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-out:
-	/* clean encrypted bundle on failure */
-	if (!res) {
-		if (g_file_test(outbundle, G_FILE_TEST_IS_REGULAR)) {
-			if (g_remove(outbundle) != 0)
-				g_warning("Failed to remove %s", outbundle);
-		}
-	}
-
-	return res;
+	g_clear_pointer(&outtmp, g_free); /* prevent removal */
+	return TRUE;
 }
 
 static gboolean is_remote_scheme(const gchar *scheme)
@@ -1753,7 +1750,6 @@ static gboolean take_bundle_ownership(int bundle_fd, GError **error)
 {
 	struct stat stat = {};
 	mode_t perm_orig = 0, perm_new = 0;
-	gboolean res = FALSE;
 	uid_t euid;
 
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
@@ -1766,8 +1762,7 @@ static gboolean take_bundle_ownership(int bundle_fd, GError **error)
 				G_FILE_ERROR,
 				g_file_error_from_errno(err),
 				"failed to fstat bundle: %s", g_strerror(err));
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
 	/* if it belongs to someone else, try to fchown if we are root */
@@ -1777,8 +1772,7 @@ static gboolean take_bundle_ownership(int bundle_fd, GError **error)
 					G_FILE_ERROR,
 					G_FILE_ERROR_PERM,
 					"cannot take file ownership of bundle when running as user (%d)", euid);
-			res = FALSE;
-			goto out;
+			return FALSE;
 		}
 
 		if (fchown(bundle_fd, 0, -1)) {
@@ -1787,8 +1781,7 @@ static gboolean take_bundle_ownership(int bundle_fd, GError **error)
 					G_FILE_ERROR,
 					g_file_error_from_errno(err),
 					"failed to chown bundle to root: %s", g_strerror(err));
-			res = FALSE;
-			goto out;
+			return FALSE;
 		}
 	}
 
@@ -1802,15 +1795,11 @@ static gboolean take_bundle_ownership(int bundle_fd, GError **error)
 					G_FILE_ERROR,
 					g_file_error_from_errno(err),
 					"failed to chmod bundle: %s", g_strerror(err));
-			res = FALSE;
-			goto out;
+			return FALSE;
 		}
 	}
 
-	res = TRUE;
-
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean check_bundle_access(int bundle_fd, GError **error)
@@ -2343,7 +2332,8 @@ gboolean check_bundle(const gchar *bundlename, RaucBundle **bundle, CheckBundleP
 			ibundle->nbd_srv->tls_ca = g_strdup(access_args->tls_ca);
 			ibundle->nbd_srv->tls_no_verify = access_args->tls_no_verify;
 			ibundle->nbd_srv->headers = g_strdupv(access_args->http_headers);
-			ibundle->nbd_srv->info_headers = g_strdupv(access_args->http_info_headers);
+			if (access_args->http_info_headers)
+				ibundle->nbd_srv->info_headers = g_ptr_array_ref(access_args->http_info_headers);
 		}
 		if (!ibundle->nbd_srv->tls_cert)
 			ibundle->nbd_srv->tls_cert = g_strdup(r_context()->config->streaming_tls_cert);
@@ -2391,8 +2381,7 @@ gboolean check_bundle(const gchar *bundlename, RaucBundle **bundle, CheckBundleP
 		gchar *path = ibundle->origpath ?: ibundle->path;
 
 		if (g_str_has_suffix(path, ".raucb")) {
-			g_autofree gchar *strprfx;
-			strprfx = g_strndup(path, strlen(path) - 6);
+			g_autofree gchar *strprfx = g_strndup(path, strlen(path) - 6);
 			ibundle->storepath = g_strconcat(strprfx, ".castr", NULL);
 		} else {
 			ibundle->storepath = g_strconcat(path, ".castr", NULL);
@@ -2651,6 +2640,8 @@ gboolean replace_signature(RaucBundle *bundle, const gchar *insig, const gchar *
 		return FALSE;
 	}
 
+	g_auto(RTempFile) outtmp = g_strdup(outpath); /* remove on early return */
+
 	if (bundle->manifest) {
 		manifest = bundle->manifest;
 	} else {
@@ -2716,14 +2707,10 @@ gboolean replace_signature(RaucBundle *bundle, const gchar *insig, const gchar *
 		goto out;
 	}
 
+	g_clear_pointer(&outtmp, g_free); /* prevent removal */
 	res = TRUE;
-out:
-	/* Remove output file on error */
-	if (!res &&
-	    g_file_test(outpath, G_FILE_TEST_IS_REGULAR))
-		if (g_remove(outpath) != 0)
-			g_warning("failed to remove %s", outpath);
 
+out:
 	/* Restore saved paths if necessary */
 	if (keyringpath || keyringdirectory) {
 		r_context()->config->keyring_path = keyringpath;
@@ -2874,16 +2861,14 @@ out:
 
 static gboolean read_complete_dm_device(gchar *dev, GError **error)
 {
-	int fd = -1;
 	const goffset chunk_size = 65536;
 	g_autofree void* buf = NULL;
 	ssize_t r;
-	gboolean ret = TRUE;
 
 	g_return_val_if_fail(dev != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	fd = g_open(dev, O_RDONLY | O_CLOEXEC, 0);
+	g_auto(filedesc) fd = g_open(dev, O_RDONLY | O_CLOEXEC, 0);
 	if (fd < 0) {
 		int err = errno;
 		g_set_error(error,
@@ -2905,14 +2890,11 @@ static gboolean read_complete_dm_device(gchar *dev, GError **error)
 					G_FILE_ERROR,
 					g_file_error_from_errno(err),
 					"Check %s device failed between %"G_GOFFSET_FORMAT " and %"G_GOFFSET_FORMAT " bytes with error: %s", dev, chunk*chunk_size, (chunk+1)*chunk_size, g_strerror(err));
-			ret = FALSE;
-			break;
+			return FALSE;
 		}
 	}
 
-	g_close(fd, NULL);
-
-	return ret;
+	return TRUE;
 }
 
 /*
@@ -3076,7 +3058,7 @@ gboolean mount_bundle(RaucBundle *bundle, GError **error)
 	GError *ierror = NULL;
 	g_autofree gchar *mount_point = NULL;
 	g_autofree gchar *loopname = NULL;
-	gint loopfd = -1;
+	gint devicefd = -1;
 	gboolean res = FALSE;
 
 	g_return_val_if_fail(bundle != NULL, FALSE);
@@ -3101,7 +3083,7 @@ gboolean mount_bundle(RaucBundle *bundle, GError **error)
 
 	if (bundle->stream) { /* local or downloaded bundle */
 		gint bundlefd = g_file_descriptor_based_get_fd(G_FILE_DESCRIPTOR_BASED(bundle->stream));
-		res = r_setup_loop(bundlefd, &loopfd, &loopname, bundle->size, &ierror);
+		res = r_setup_loop(bundlefd, &devicefd, &loopname, bundle->size, &ierror);
 		if (!res) {
 			g_propagate_error(error, ierror);
 			goto out;
@@ -3111,7 +3093,7 @@ gboolean mount_bundle(RaucBundle *bundle, GError **error)
 		bundle->nbd_dev->data_size = bundle->size;
 		bundle->nbd_dev->sock = bundle->nbd_srv->sock;
 		bundle->nbd_srv->sock = -1;
-		res = r_nbd_setup_device(bundle->nbd_dev, &ierror);
+		res = r_nbd_setup_device(bundle->nbd_dev, &devicefd, &ierror);
 		if (!res) {
 			/* The setup failed, so the socket still belongs to the nbd_srv. */
 			bundle->nbd_srv->sock = bundle->nbd_dev->sock;
@@ -3193,49 +3175,43 @@ out:
 	if (mount_point) {
 		g_rmdir(mount_point);
 	}
-	if (loopfd >= 0)
-		g_close(loopfd, NULL);
+	if (devicefd >= 0)
+		g_close(devicefd, NULL);
 	return res;
 }
 
 gboolean umount_bundle(RaucBundle *bundle, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	g_return_val_if_fail(bundle != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
 	g_assert_nonnull(bundle->mount_point);
 
-	res = r_umount_bundle(bundle->mount_point, &ierror);
-	if (!res) {
+	if (!r_umount_bundle(bundle->mount_point, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	g_rmdir(bundle->mount_point);
 	g_clear_pointer(&bundle->mount_point, g_free);
 
 	if (ENABLE_STREAMING && bundle->nbd_dev) {
-		res = r_nbd_remove_device(bundle->nbd_dev, &ierror);
-		if (!res) {
+		if (!r_nbd_remove_device(bundle->nbd_dev, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	if (ENABLE_STREAMING && bundle->nbd_srv) {
-		res = r_nbd_stop_server(bundle->nbd_srv, &ierror);
-		if (!res) {
+		if (!r_nbd_stop_server(bundle->nbd_srv, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
-	res = TRUE;
-out:
-	return res;
+	return TRUE;
 }
 
 void free_bundle(RaucBundle *bundle)
@@ -3276,6 +3252,92 @@ void free_bundle(RaucBundle *bundle)
 	g_free(bundle);
 }
 
+static gchar* get_uptime(void)
+{
+	g_autofree gchar *contents = NULL;
+	g_autoptr(GError) ierror = NULL;
+	g_auto(GStrv) uptime = NULL;
+
+	if (!g_file_get_contents("/proc/uptime", &contents, NULL, &ierror)) {
+		g_warning("Failed to get uptime: %s", ierror->message);
+		return NULL;
+	}
+
+	/* file contains two values and a newline, 'chomp' in-place and split then */
+	uptime = g_strsplit(g_strchomp(contents), " ", 2);
+
+	return g_strdup(uptime[0]);
+}
+
+/* If the input key starts with RAUC_HTTP_, it returns a valid HTTP header
+ * string with 'RAUC_HTTP_' replaced by 'RAUC-'.
+ * If the input string does not start with RAUC_HTTP_, NULL is returned.
+ */
+static gchar *system_info_to_header(const gchar *key, const gchar *value)
+{
+	g_autofree gchar *header_key = NULL;
+
+	g_return_val_if_fail(key, NULL);
+	g_return_val_if_fail(value, NULL);
+
+	if (!g_str_has_prefix(key, "RAUC_HTTP_"))
+		return NULL;
+
+	header_key = g_strdup(key + strlen("RAUC_HTTP_"));
+	for (size_t i = 0; i < strlen(header_key); i++) {
+		if (header_key[i] == '_')
+			header_key[i] = '-';
+	}
+
+	return g_strdup_printf("RAUC-%s: %s", header_key, value);
+}
+
+GPtrArray *assemble_info_headers(const gchar *transaction)
+{
+	g_autoptr(GPtrArray) headers = g_ptr_array_new_with_free_func(g_free);
+
+	if (!r_context()->config->enabled_headers)
+		goto no_std_headers;
+
+	for (gchar **header = r_context()->config->enabled_headers; *header; header++) {
+		/* Add static system information */
+		if (g_strcmp0(*header, "boot-id") == 0 && r_context()->boot_id)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Boot-ID: %s", r_context()->boot_id));
+		if (g_strcmp0(*header, "machine-id") == 0 && r_context()->machine_id)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Machine-ID: %s", r_context()->machine_id));
+		if (g_strcmp0(*header, "system-version") == 0 && r_context()->system_version)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-System-Version: %s", r_context()->system_version));
+		if (g_strcmp0(*header, "serial") == 0 && r_context()->system_serial)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Serial: %s", r_context()->system_serial));
+		if (g_strcmp0(*header, "variant") == 0 && r_context()->config->system_variant)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Variant: %s", r_context()->config->system_variant));
+		/* Add per-installation information */
+		if (g_strcmp0(*header, "transaction-id") == 0 && transaction)
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Transaction-ID: %s", transaction));
+		/* Add live information */
+		if (g_strcmp0(*header, "uptime") == 0) {
+			g_autofree gchar *uptime = get_uptime();
+			g_ptr_array_add(headers, g_strdup_printf("RAUC-Uptime: %s", uptime));
+		}
+	}
+
+no_std_headers:
+
+	if (r_context()->system_info) {
+		GHashTableIter iter;
+		gchar *key = NULL;
+		gchar *value = NULL;
+
+		g_hash_table_iter_init(&iter, r_context()->system_info);
+		while (g_hash_table_iter_next(&iter, (gpointer*) &key, (gpointer*) &value)) {
+			gchar *header = system_info_to_header(key, value);
+			if (header)
+				g_ptr_array_add(headers, header);
+		}
+	}
+
+	return g_steal_pointer(&headers);
+}
 void clear_bundle_access_args(RaucBundleAccessArgs *access_args)
 {
 	if (ENABLE_STREAMING) {
@@ -3283,7 +3345,7 @@ void clear_bundle_access_args(RaucBundleAccessArgs *access_args)
 		g_free(access_args->tls_key);
 		g_free(access_args->tls_ca);
 		g_strfreev(access_args->http_headers);
-		g_strfreev(access_args->http_info_headers);
+		g_clear_pointer(&access_args->http_info_headers, g_ptr_array_unref);
 	}
 
 	memset(access_args, 0, sizeof(*access_args));
