@@ -16,21 +16,6 @@
 RaucContext *context = NULL;
 gboolean context_configuring = FALSE;
 
-static gchar *regex_match(const gchar *pattern, const gchar *string)
-{
-	g_autoptr(GRegex) regex = NULL;
-	g_autoptr(GMatchInfo) match = NULL;
-
-	g_return_val_if_fail(pattern, NULL);
-	g_return_val_if_fail(string, NULL);
-
-	regex = g_regex_new(pattern, 0, 0, NULL);
-	if (g_regex_match(regex, string, 0, &match))
-		return g_match_info_fetch(match, 1);
-
-	return NULL;
-}
-
 static gchar* get_machine_id(void)
 {
 	gchar *contents = NULL;
@@ -42,7 +27,13 @@ static gchar* get_machine_id(void)
 	}
 
 	/* file contains newline, modify in-place */
-	return g_strchomp(contents);
+	g_strchomp(contents);
+	if (!contents[0]) {
+		g_info("Failed to get machine-id: empty file");
+		return NULL;
+	}
+
+	return contents;
 }
 
 static gchar* get_boot_id(void)
@@ -59,44 +50,57 @@ static gchar* get_boot_id(void)
 	return g_strchomp(contents);
 }
 
-static gchar* get_cmdline_bootname(void)
+static gchar *get_cmdline(void)
 {
-	g_autofree gchar *contents = NULL;
-	g_autofree gchar *realdev = NULL;
-	gchar *bootname = NULL;
+	gchar *contents = NULL;
+	g_autoptr(GError) ierror = NULL;
 
 	if (context->mock.proc_cmdline)
-		contents = g_strdup(context->mock.proc_cmdline);
-	else if (!g_file_get_contents("/proc/cmdline", &contents, NULL, NULL))
+		return g_strdup(context->mock.proc_cmdline);
+
+	if (!g_file_get_contents("/proc/cmdline", &contents, NULL, &ierror)) {
+		g_warning("Failed to get kernel cmdline: %s", ierror->message);
+		return NULL;
+	}
+
+	return contents;
+}
+
+static gchar* get_cmdline_bootname_explicit(const gchar *cmdline)
+{
+	if (!cmdline)
 		return NULL;
 
-	if (strstr(contents, "rauc.external") != NULL) {
+	gchar *bootname = NULL;
+
+	if (strstr(cmdline, "rauc.external") != NULL) {
 		g_message("Detected explicit external boot, ignoring missing active slot");
 		return g_strdup("_external_");
 	}
 
-	bootname = regex_match("rauc\\.slot=(\\S+)", contents);
+	if (strstr(cmdline, "root=/dev/nfs") != NULL) {
+		g_message("Detected nfs boot, ignoring missing active slot");
+		return g_strdup("_external_");
+	}
+
+	bootname = r_regex_match_simple("rauc\\.slot=(\\S+)", cmdline);
 	if (bootname)
 		return bootname;
 
-	/* For barebox, we check if the bootstate code set the active slot name
-	 * in the command line */
-	if (g_strcmp0(context->config->system_bootloader, "barebox") == 0) {
-		bootname = regex_match(
-				"(?:bootstate|bootchooser)\\.active=(\\S+)",
-				contents);
-		if (bootname)
-			return bootname;
-	}
+	return NULL;
+}
 
-	bootname = regex_match("root=(\\S+)", contents);
-	if (g_strcmp0(bootname, "/dev/nfs") == 0) {
-		g_message("Detected nfs boot, ignoring missing active slot");
-		g_free(bootname);
-		return g_strdup("_external_");
-	}
+static gchar* get_cmdline_bootname_root(const gchar *cmdline)
+{
+	if (!cmdline)
+		return NULL;
+
+	g_autofree gchar *realdev = NULL;
+	gchar *bootname = NULL;
+
+	bootname = r_regex_match_simple("root=(\\S+)", cmdline);
 	if (!bootname)
-		bootname = regex_match("systemd\\.verity_root_data=(\\S+)", contents);
+		bootname = r_regex_match_simple("systemd\\.verity_root_data=(\\S+)", cmdline);
 
 	if (!bootname)
 		return NULL;
@@ -148,6 +152,29 @@ static gchar* get_cmdline_bootname(void)
 	return bootname;
 }
 
+static gchar* get_bootname(void)
+{
+	g_autofree gchar *cmdline = get_cmdline();
+	gchar *bootname = NULL;
+	GError *ierror = NULL;
+
+	bootname = get_cmdline_bootname_explicit(cmdline);
+	if (bootname)
+		return bootname;
+
+	bootname = r_boot_get_current_bootname(context->config, cmdline, &ierror);
+	if (ierror) {
+		g_message("Failed to get bootname: %s", ierror->message);
+		g_clear_error(&ierror);
+	}
+	if (bootname)
+		return bootname;
+
+	bootname = get_cmdline_bootname_root(cmdline);
+
+	return bootname;
+}
+
 /**
  * Launches a handler and obtains variables from output by looking for
  * 'RAUC_<SOMETHING>=value' lines to put them into a key/value store (GHashTable).
@@ -323,7 +350,7 @@ static gboolean r_context_configure_target(GError **error)
 		}
 	}
 
-	/* load system status if central status file is available */
+	/* load system status and slot status if central status file is available */
 	if (g_strcmp0(context->config->statusfile_path, "per-slot") != 0) {
 		g_clear_pointer(&context->system_status, r_system_status_free);
 		context->system_status = g_new0(RSystemStatus, 1);
@@ -331,6 +358,8 @@ static gboolean r_context_configure_target(GError **error)
 			g_message("Failed to load system status: %s", ierror->message);
 			g_clear_error(&ierror);
 		}
+
+		r_slot_status_load_globally(context->config->statusfile_path, context->config->slots);
 	}
 
 	/* set up logging */
@@ -374,15 +403,7 @@ static gboolean r_context_configure_target(GError **error)
 		g_warning("Ignoring surrounding whitespace in system variant: %s", context->config->system_variant);
 
 	if (context->bootslot == NULL) {
-		context->bootslot = get_cmdline_bootname();
-	}
-
-	if (context->bootslot == NULL) {
-		context->bootslot = r_boot_get_current_bootname(context->config, &ierror);
-		if (ierror) {
-			g_message("Failed to get bootname: %s", ierror->message);
-			g_clear_error(&ierror);
-		}
+		context->bootslot = get_bootname();
 	}
 
 	g_clear_pointer(&context->boot_id, g_free);
@@ -464,7 +485,10 @@ gboolean r_context_configure(GError **error)
 		/* This is a hack as we cannot get rid of config easily */
 		/* Fallthrough */
 		case R_CONTEXT_CONFIG_MODE_NONE:
-			default_config(&context->config);
+			if (!default_config(&context->config, &ierror)) {
+				g_propagate_error(error, ierror);
+				return FALSE;
+			}
 			break;
 		default:
 			g_error("invalid context config mode %d", configmode);
@@ -603,7 +627,8 @@ void r_context_begin_step_weighted(const gchar *name, const gchar *description,
 	/* calculate percentage */
 	if (context->progress) {
 		parent = context->progress->data;
-		g_assert_cmpint(parent->substeps_total, >, 0);
+		if (parent->substeps_total == 0)
+			g_error("Cannot add substep '%s': Parent step '%s' has no substeps.", name, parent->name);
 
 		/* nesting check */
 		if (parent->substeps_total == parent->substeps_done)
@@ -730,13 +755,17 @@ void r_context_set_step_percentage(const gchar *name, gint custom_percent)
 	g_assert_nonnull(context->progress);
 
 	step = context->progress->data;
+	if (!g_list_next(context->progress))
+		g_error("Root step does not support setting percentage");
 	parent = g_list_next(context->progress)->data;
 
 	/* ensure that progress step nesting is done correctly */
-	g_assert_cmpstr(step->name, ==, name);
+	if (g_strcmp0(step->name, name) != 0)
+		g_error("Wrong step context '%s'. Current step is '%s'.", name, step->name);
 
 	/* substeps and setting explicit percentage does not make sense */
-	g_assert_cmpint(step->substeps_total, ==, 0);
+	if (step->substeps_total < 0)
+		g_error("Setting percentage on substeps > 0 is not supported");
 
 	percent_difference = custom_percent - step->last_explicit_percent;
 
@@ -759,6 +788,12 @@ void r_context_set_step_percentage(const gchar *name, gint custom_percent)
 		r_context_send_progress(FALSE, FALSE);
 }
 
+void r_context_inc_step_percentage(const gchar *name)
+{
+	RaucProgressStep *step = context->progress->data;
+	r_context_set_step_percentage(name, step->last_explicit_percent + 1);
+}
+
 void r_context_free_progress_step(RaucProgressStep *step)
 {
 	if (!step)
@@ -780,7 +815,9 @@ void r_context_register_progress_callback(progress_callback progress_cb)
 
 RaucContext *r_context_conf(void)
 {
-	if (context == NULL) {
+	static gboolean initialized = FALSE;
+
+	if (!initialized) {
 		GError *ierror = NULL;
 
 		// let us handle broken pipes explicitly
@@ -797,6 +834,10 @@ RaucContext *r_context_conf(void)
 			return NULL;
 		}
 
+		initialized = TRUE;
+	}
+
+	if (context == NULL) {
 		context = g_new0(RaucContext, 1);
 		context->progress = NULL;
 		context->install_info = g_new0(RContextInstallationInfo, 1);
@@ -813,7 +854,8 @@ const RaucContext *r_context(void)
 {
 	GError *ierror = NULL;
 
-	g_assert_nonnull(context);
+	if (!context)
+		g_error("Context not initialized. Call r_context_conf() first");
 
 	if (context_configuring)
 		g_error("Detected call of r_context() while still setting up context! Aborted to avoid infinite recursion!");
