@@ -4,17 +4,64 @@
 #include "config_file.h"
 #include "context.h"
 #include "manifest.h"
+#include "glib.h"
 #include "signature.h"
+#include "update_handler.h"
 #include "utils.h"
 
 #define RAUC_IMAGE_PREFIX	"image"
 
 #define R_MANIFEST_ERROR r_manifest_error_quark()
+
 GQuark r_manifest_error_quark(void)
 {
 	return g_quark_from_static_string("r_manifest_error_quark");
 }
 
+static gboolean handle_missing_type(RaucImage *image, GError **error)
+{
+	/* When using custom install hooks, having no type is okay */
+	if (image->hooks.install)
+		return TRUE;
+
+	const gchar *derived_type = derive_image_type_from_filename_pattern(image->filename);
+	if (!derived_type) {
+		g_set_error(error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
+				"No 'type=' set and unable to map extension of file '%s' to known image type",
+				image->filename);
+		return FALSE;
+	}
+
+	image->type_from_fileext = TRUE;
+	image->type = g_strdup(derived_type);
+	return TRUE;
+}
+
+static gboolean validate_filename_requirements(RaucImage *image, GError **error)
+{
+	gboolean has_filename = (image->filename != NULL);
+	gboolean has_install_hook = image->hooks.install;
+
+	/* Combining 'type=emptyfs' with an image filename would be contradictory */
+	if (g_strcmp0(image->type, "emptyfs") == 0) {
+		if (has_filename) {
+			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_PARSE_ERROR,
+					"It is not supported setting 'filename' when 'type=emptyfs' is set");
+			return FALSE;
+		}
+	} else {
+		/* All other image types require either a source file
+		 * or custom install hooks. */
+		if (!has_filename && !has_install_hook) {
+			g_set_error(error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND,
+					"Missing required 'filename'");
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
 static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **image, GError **error)
 {
 	g_autoptr(RaucImage) iimage = r_new_image();
@@ -24,7 +71,6 @@ static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **
 	gsize entries;
 	g_auto(GStrv) converted = NULL;
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	g_return_val_if_fail(key_file != NULL, FALSE);
 	g_return_val_if_fail(group != NULL, FALSE);
@@ -67,7 +113,7 @@ static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **
 		g_clear_error(&ierror);
 	} else if (ierror) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 	g_key_file_remove_key(key_file, group, "size", NULL);
 
@@ -82,20 +128,54 @@ static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **
 		} else {
 			g_set_error(error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_PARSE,
 					"slot hook type '%s' not supported", hooks[j]);
-			goto out;
+			return FALSE;
 		}
 	}
 	g_key_file_remove_key(key_file, group, "hooks", NULL);
 
 	iimage->filename = key_file_consume_string(key_file, group, "filename", &ierror);
-	/* 'filename' is optional only for 'install' hooks */
-	if (iimage->filename == NULL) {
-		if (!iimage->hooks.install) {
-			g_propagate_error(error, ierror);
-			goto out;
-		} else {
+	/* A missing 'filename' can be correct, as it is optional for 'install' hooks and 'type=emptyfs'.
+	 * So we collect all requirements first and check their validity afterwards */
+	if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
+		g_clear_error(&ierror);
+	} else if (ierror) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
+	/* Setting the 'type' option for artifacts is not supported.
+	 * For regular images (non-artifacts), we need to determine the image 'type'
+	 * to select the appropriate update handler. The image 'type' can be determined either
+	 * by the corresponding variable in the manifest, or derived by the file name extension. */
+	if (!iimage->artifact) {
+		iimage->type_from_fileext = FALSE;
+		iimage->type = key_file_consume_string(key_file, group, "type", &ierror);
+		if (g_error_matches(ierror, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
 			g_clear_error(&ierror);
+			g_clear_pointer(&iimage->type, g_free);
+			if (!handle_missing_type(iimage, &ierror)) {
+				g_propagate_error(error, ierror);
+				return FALSE;
+			}
+		} else if (ierror) {
+			g_propagate_error(error, ierror);
+			return FALSE;
 		}
+
+		/* Custom install hooks can skip validation of supported image types
+		 * since they implement their own logic */
+		if (!iimage->hooks.install && !is_image_type_supported(iimage->type)) {
+			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_ERROR_INVALID_IMAGE_TYPE,
+					"Unsupported image type '%s'", iimage->type);
+			return FALSE;
+		}
+	}
+
+	/* All requirements to check if a filename is necessary have been collected,
+	 * so we can now check if the current state is valid */
+	if (!validate_filename_requirements(iimage, &ierror)) {
+		g_propagate_error(error, ierror);
+		return FALSE;
 	}
 
 	g_key_file_remove_key(key_file, group, "version", NULL);
@@ -117,15 +197,13 @@ static gboolean parse_image(GKeyFile *key_file, const gchar *group, RaucImage **
 
 	if (!check_remaining_keys(key_file, group, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 	g_key_file_remove_group(key_file, group, NULL);
 
-	res = TRUE;
 	*image = g_steal_pointer(&iimage);
 
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean parse_meta(GKeyFile *key_file, const gchar *group, RaucManifest *raucm, GError **error)
@@ -223,6 +301,7 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 	raucm->update_version = key_file_consume_string(key_file, "update", "version", NULL);
 	raucm->update_description = key_file_consume_string(key_file, "update", "description", NULL);
 	raucm->update_build = key_file_consume_string(key_file, "update", "build", NULL);
+	raucm->update_min_rauc_version = key_file_consume_string(key_file, "update", "min-rauc-version", NULL);
 	if (!check_remaining_keys(key_file, "update", &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -311,7 +390,7 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 			RaucImage *image = NULL;
 
 			if (!parse_image(key_file, groups[i], &image, &ierror)) {
-				g_propagate_error(error, ierror);
+				g_propagate_prefixed_error(error, ierror, "Cannot parse [%s]: ", groups[i]);
 				return FALSE;
 			}
 
@@ -326,6 +405,10 @@ static gboolean parse_manifest(GKeyFile *key_file, RaucManifest **manifest, GErr
 		}
 	}
 
+	/* ignore [rollout] section for now, so that we can add hints/overrides
+	 * for rollout/polling behaviour later */
+	g_key_file_remove_group(key_file, "rollout", NULL);
+
 	if (!check_remaining_groups(key_file, &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -344,6 +427,10 @@ gboolean load_manifest_mem(GBytes *mem, RaucManifest **manifest, GError **error)
 	gsize length;
 	g_autofree gchar *manifest_checksum = NULL;
 
+	g_return_val_if_fail(mem, FALSE);
+	g_return_val_if_fail(manifest != NULL && *manifest == NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	data = g_bytes_get_data(mem, &length);
 	if (data == NULL) {
 		g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_ERROR_NO_DATA, "No data available");
@@ -372,11 +459,15 @@ gboolean load_manifest_mem(GBytes *mem, RaucManifest **manifest, GError **error)
 gboolean load_manifest_file(const gchar *filename, RaucManifest **manifest, GError **error)
 {
 	GError *ierror = NULL;
-	g_autofree gchar *data;
+	g_autofree gchar *data = NULL;
 	gsize length;
 	g_autoptr(GKeyFile) key_file = NULL;
 	g_autofree gchar *manifest_checksum = NULL;
 
+	g_return_val_if_fail(filename, FALSE);
+	g_return_val_if_fail(manifest != NULL && *manifest == NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	if (!g_file_get_contents(filename, &data, &length, &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -403,7 +494,33 @@ gboolean load_manifest_file(const gchar *filename, RaucManifest **manifest, GErr
 static gboolean check_manifest_common(const RaucManifest *mf, GError **error)
 {
 	gboolean have_hooks = FALSE;
-	gboolean res = FALSE;
+
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	if (mf->update_min_rauc_version) {
+		if (!r_semver_less_equal("0", mf->update_min_rauc_version, NULL)) {
+			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR,
+					"Failed to parse 'min-rauc-version'. Expected 'Major[.Minor[.Patch]][-pre_release]]', got '%s'",
+					mf->update_min_rauc_version
+					);
+			return FALSE;
+		}
+		if (!r_semver_less_equal(mf->update_min_rauc_version, RAUC_MESON_VERSION, NULL)) {
+			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR,
+					"Minimum RAUC version in manifest (%s) is newer than current version (%s)",
+					mf->update_min_rauc_version, RAUC_MESON_VERSION
+					);
+			return FALSE;
+		}
+		if (r_semver_less_equal(mf->update_min_rauc_version, "1.13", NULL)) {
+			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR,
+					"Minimum RAUC version field in manifest is only supported since 1.14 (not '%s')",
+					mf->update_min_rauc_version
+					);
+			return FALSE;
+		}
+	}
 
 	switch (mf->bundle_format) {
 		case R_MANIFEST_FORMAT_PLAIN:
@@ -413,7 +530,7 @@ static gboolean check_manifest_common(const RaucManifest *mf, GError **error)
 			break; /* data checked in _detached/_inline */
 		default: {
 			g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Unsupported bundle format");
-			goto out;
+			return FALSE;
 		}
 	}
 
@@ -440,12 +557,10 @@ static gboolean check_manifest_common(const RaucManifest *mf, GError **error)
 
 	if (have_hooks && !mf->hook_name) {
 		g_set_error(error, R_MANIFEST_ERROR, R_MANIFEST_CHECK_ERROR, "Hooks used, but no hook 'filename' defined in [hooks] section");
-		goto out;
+		return FALSE;
 	}
 
-	res = TRUE;
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean check_manifest_plain(const RaucManifest *mf, GError **error)
@@ -480,6 +595,9 @@ static gboolean check_manifest_plain(const RaucManifest *mf, GError **error)
  */
 static gboolean check_manifest_bundled(const RaucManifest *mf, GError **error)
 {
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	for (GList *l = mf->images; l != NULL; l = l->next) {
 		RaucImage *image = l->data;
 
@@ -536,6 +654,9 @@ gboolean check_manifest_input(const RaucManifest *mf, GError **error)
 {
 	GError *ierror = NULL;
 
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	if (!check_manifest_common(mf, &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -594,6 +715,9 @@ gboolean check_manifest_internal(const RaucManifest *mf, GError **error)
 	GError *ierror = NULL;
 	gboolean res = FALSE;
 
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	r_context_begin_step("check_manifest", "Checking manifest contents", 0);
 
 	if (!check_manifest_common(mf, &ierror)) {
@@ -651,6 +775,9 @@ gboolean check_manifest_external(const RaucManifest *mf, GError **error)
 	gboolean res = FALSE;
 	const gchar *format = NULL;
 
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	r_context_begin_step("check_manifest", "Checking manifest contents", 0);
 
 	format = r_manifest_bundle_format_to_str(mf->bundle_format);
@@ -741,6 +868,9 @@ out:
 
 gboolean check_manifest_create(const RaucManifest *mf, GError **error)
 {
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	for (GList *l = mf->images; l != NULL; l = l->next) {
 		RaucImage *image = l->data;
 
@@ -764,6 +894,8 @@ static GKeyFile *prepare_manifest(const RaucManifest *mf)
 	GHashTable *meta_kvs;
 	const gchar *meta_group;
 
+	g_return_val_if_fail(mf, FALSE);
+
 	key_file = g_key_file_new();
 
 	if (mf->update_compatible)
@@ -852,6 +984,9 @@ static GKeyFile *prepare_manifest(const RaucManifest *mf)
 		if (image->filename)
 			g_key_file_set_string(key_file, group, "filename", image->filename);
 
+		if (image->type && !image->type_from_fileext)
+			g_key_file_set_string(key_file, group, "type", image->type);
+
 		if (image->hooks.pre_install == TRUE) {
 			g_ptr_array_add(hooklist, g_strdup("pre-install"));
 		}
@@ -924,6 +1059,10 @@ gboolean save_manifest_file(const gchar *filename, const RaucManifest *mf, GErro
 	g_autoptr(GKeyFile) key_file = NULL;
 	gboolean res = FALSE;
 
+	g_return_val_if_fail(filename, FALSE);
+	g_return_val_if_fail(mf, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	key_file = prepare_manifest(mf);
 
 	res = g_key_file_save_to_file(key_file, filename, &ierror);
@@ -942,6 +1081,8 @@ GVariant* r_manifest_to_dict(const RaucManifest *manifest)
 	GHashTable *kvs;
 	const gchar *group;
 
+	g_return_val_if_fail(manifest, NULL);
+
 	g_variant_dict_init(&root_dict, NULL);
 
 	if (manifest->hash)
@@ -1007,6 +1148,8 @@ GVariant* r_manifest_to_dict(const RaucManifest *manifest)
 			g_variant_builder_add(&builder, "{sv}", "variant", g_variant_new_string(img->variant));
 		if (img->filename)
 			g_variant_builder_add(&builder, "{sv}", "filename", g_variant_new_string(img->filename));
+		if (img->type)
+			g_variant_builder_add(&builder, "{sv}", "type", g_variant_new_string(img->type));
 		if (img->checksum.digest)
 			g_variant_builder_add(&builder, "{sv}", "checksum", g_variant_new_string(img->checksum.digest));
 		if (img->checksum.size)
@@ -1108,9 +1251,10 @@ void r_free_image(gpointer data)
 	g_free(image->variant);
 	g_free(image->checksum.digest);
 	g_free(image->filename);
+	g_free(image->type);
 	g_strfreev(image->adaptive);
 	g_strfreev(image->convert);
-	g_clear_pointer(&image->converted, (GDestroyNotify)g_ptr_array_unref);
+	g_clear_pointer(&image->converted, g_ptr_array_unref);
 	g_free(image);
 }
 
@@ -1123,6 +1267,7 @@ void free_manifest(RaucManifest *manifest)
 	g_free(manifest->update_version);
 	g_free(manifest->update_description);
 	g_free(manifest->update_build);
+	g_free(manifest->update_min_rauc_version);
 	g_free(manifest->bundle_verity_hash);
 	g_free(manifest->bundle_verity_salt);
 	g_free(manifest->bundle_crypt_key);
@@ -1130,9 +1275,9 @@ void free_manifest(RaucManifest *manifest)
 	g_free(manifest->handler_args);
 	g_free(manifest->hook_name);
 	g_list_free_full(manifest->images, r_free_image);
-	g_clear_pointer(&manifest->meta, (GDestroyNotify)g_hash_table_destroy);
+	g_clear_pointer(&manifest->meta, g_hash_table_destroy);
 	g_free(manifest->hash);
-	g_clear_pointer(&manifest->warnings, (GDestroyNotify)g_ptr_array_unref);
+	g_clear_pointer(&manifest->warnings, g_ptr_array_unref);
 	g_free(manifest);
 }
 
@@ -1152,6 +1297,10 @@ static gboolean update_manifest_checksums(RaucManifest *manifest, const gchar *d
 	gboolean res = TRUE;
 	gboolean had_errors = FALSE;
 
+	g_return_val_if_fail(manifest, FALSE);
+	g_return_val_if_fail(dir, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
 	for (GList *elem = manifest->images; elem != NULL; elem = elem->next) {
 		RaucImage *image = elem->data;
 		g_autofree gchar *filename = NULL;
