@@ -5,13 +5,27 @@
 #include <openssl/evp.h>
 #include <openssl/pem.h>
 #include <openssl/crypto.h>
+#if ENABLE_OPENSSL_PKCS11_ENGINE
 #include <openssl/engine.h>
+#endif
 #include <openssl/x509.h>
 #include <string.h>
 
 #include "context.h"
 #include "signature.h"
 
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(X509_STORE_CTX, X509_STORE_CTX_free);
+
+void r_signature_free_x509_stack(R_X509_STACK *stack)
+{
+	sk_X509_free(stack);
+}
+
+void r_signature_free_x509_stack_pop(R_X509_STACK_POP *stack)
+{
+	sk_X509_pop_free(stack, X509_free);
+}
+
 GQuark r_signature_error_quark(void)
 {
 	return g_quark_from_static_string("r_signature_error_quark");
@@ -93,14 +107,12 @@ gboolean signature_init(GError **error)
 		return FALSE;
 	}
 
-	id = X509_PURPOSE_get_count() + 1;
-	if (X509_PURPOSE_get_by_id(id) >= 0) {
-		g_set_error_literal(
-				error,
-				R_SIGNATURE_ERROR,
-				R_SIGNATURE_ERROR_CRYPTOINIT_FAILED,
-				"Failed to calculate free OpenSSL X509 purpose id");
-		return FALSE;
+	/* OpenSSL 3.5 warns that there may be gaps, so we need to search.
+	 * When we have 3.5 as the minimum version, we can use
+	 * X509_PURPOSE_get_unused_id instead. */
+	id = X509_PURPOSE_MAX + 1;
+	while (X509_PURPOSE_get_by_id(id) != -1) {
+		id++;
 	}
 
 	/* X509_TRUST_OBJECT_SIGN maps to the Code Signing ID (via OpenSSL's NID_code_sign) */
@@ -117,6 +129,7 @@ gboolean signature_init(GError **error)
 	return TRUE;
 }
 
+#if ENABLE_OPENSSL_PKCS11_ENGINE
 static ENGINE *get_pkcs11_engine(GError **error)
 {
 	static ENGINE *e = NULL;
@@ -181,6 +194,7 @@ free:
 out:
 	return e;
 }
+#endif
 
 static EVP_PKEY *load_key_file(const gchar *keyfile, GError **error)
 {
@@ -221,6 +235,7 @@ out:
 static EVP_PKEY *load_key_pkcs11(const gchar *url, GError **error)
 {
 	EVP_PKEY *res = NULL;
+#if ENABLE_OPENSSL_PKCS11_ENGINE
 	GError *ierror = NULL;
 	ENGINE *e;
 
@@ -242,6 +257,14 @@ static EVP_PKEY *load_key_pkcs11(const gchar *url, GError **error)
 				"failed to load PKCS11 private key for '%s': %s", url, get_openssl_err_string());
 		goto out;
 	}
+#else
+	g_set_error(
+			error,
+			R_SIGNATURE_ERROR,
+			R_SIGNATURE_ERROR_LOAD_FAILED,
+			"failed to load PKCS11 private key for '%s': OpenSSL engine support disabled", url);
+#endif
+
 out:
 	return res;
 }
@@ -293,7 +316,7 @@ static STACK_OF(X509) *load_certs_from_file(const gchar *certfile, GError **erro
 {
 	BIO *cert_bio = NULL;
 	X509 *cert_x509 = NULL;
-	STACK_OF(X509) *certs = NULL;
+	g_autoptr(R_X509_STACK_POP) certs = NULL;
 	unsigned long err;
 
 	g_return_val_if_fail(certfile != NULL, NULL);
@@ -329,8 +352,7 @@ static STACK_OF(X509) *load_certs_from_file(const gchar *certfile, GError **erro
 					R_SIGNATURE_ERROR_PARSE_ERROR,
 					"Failed to parse cert file '%s': %s", certfile, get_openssl_err_string());
 			/* other certs loaded so far are not required anymore and must be freed */
-			sk_X509_pop_free(certs, X509_free);
-			certs = NULL;
+			g_clear_pointer(&certs, r_signature_free_x509_stack_pop);
 			goto out;
 		}
 
@@ -339,12 +361,13 @@ static STACK_OF(X509) *load_certs_from_file(const gchar *certfile, GError **erro
 
 out:
 	BIO_free_all(cert_bio);
-	return certs;
+	return g_steal_pointer(&certs);
 }
 
 static X509 *load_cert_pkcs11(const gchar *url, GError **error)
 {
 	X509 *res = NULL;
+#if ENABLE_OPENSSL_PKCS11_ENGINE
 	GError *ierror = NULL;
 	ENGINE *e;
 
@@ -374,6 +397,13 @@ static X509 *load_cert_pkcs11(const gchar *url, GError **error)
 		goto out;
 	}
 	res = parms.cert;
+#else
+	g_set_error(
+			error,
+			R_SIGNATURE_ERROR,
+			R_SIGNATURE_ERROR_PARSE_ERROR,
+			"failed to load PKCS11 certificate for '%s': OpenSSL engine support disabled", url);
+#endif
 
 out:
 	return res;
@@ -524,7 +554,7 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 	BIO *outsig = BIO_new(BIO_s_mem());
 	g_autoptr(X509) signcert = NULL;
 	g_autoptr(EVP_PKEY) pkey = NULL;
-	STACK_OF(X509) *intercerts = NULL;
+	g_autoptr(R_X509_STACK_POP) intercerts = NULL;
 	g_autoptr(CMS_ContentInfo) cms = NULL;
 	GBytes *res = NULL;
 	int flags = CMS_BINARY | CMS_NOSMIMECAP;
@@ -603,7 +633,7 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 	if (keyring_path || keyring_dir) {
 		g_autoptr(CMS_ContentInfo) vcms = NULL;
 		g_autoptr(X509_STORE) store = NULL;
-		STACK_OF(X509) *verified_chain = NULL;
+		g_autoptr(R_X509_STACK_POP) verified_chain = NULL;
 		g_autoptr(GBytes) manifest = NULL;
 
 		if (!(store = setup_x509_store(keyring_path, keyring_dir, &ierror))) {
@@ -647,20 +677,112 @@ GBytes *cms_sign(GBytes *content, gboolean detached, const gchar *certfile, cons
 				g_warning("Certificate %d (%s) will expire in less than a month!", i + 1, buf);
 			}
 		}
-
-		sk_X509_pop_free(verified_chain, X509_free);
 	} else {
 		g_message("No keyring given, skipping signature verification");
 	}
 out:
 	ERR_print_errors_fp(stdout);
-	if (intercerts)
-		sk_X509_pop_free(intercerts, X509_free);
 	BIO_free_all(incontent);
 	BIO_free_all(outsig);
 	return res;
 }
 
+static void debug_cms_ci(CMS_ContentInfo *cms);
+
+GBytes *cms_append_signature(GBytes *input_sig, const gchar *certfile, const gchar *keyfile, gchar **interfiles, GError **error)
+{
+	GError *ierror = NULL;
+	g_autoptr(CMS_ContentInfo) cms = NULL;
+	BIO *insig = bytes_as_bio(input_sig);
+	BIO *outsig = BIO_new(BIO_s_mem());
+	g_autoptr(X509) signcert = NULL;
+	g_autoptr(EVP_PKEY) pkey = NULL;
+	GBytes *output_sig = NULL;
+	int flags = CMS_BINARY | CMS_NOSMIMECAP | CMS_REUSE_DIGEST;
+
+	g_return_val_if_fail(input_sig != NULL, NULL);
+	g_return_val_if_fail(certfile != NULL, NULL);
+	g_return_val_if_fail(keyfile != NULL, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	if (!(cms = d2i_CMS_bio(insig, NULL))) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_PARSE,
+				"failed to parse signature");
+		goto out;
+	}
+
+	debug_cms_ci(cms);
+
+	signcert = load_cert(certfile, &ierror);
+	if (signcert == NULL) {
+		g_propagate_error(error, ierror);
+		goto out;
+	}
+
+	pkey = load_key(keyfile, &ierror);
+	if (pkey == NULL) {
+		g_propagate_error(error, ierror);
+		goto out;
+	}
+
+	for (gchar **intercertpath = interfiles; intercertpath && *intercertpath != NULL; intercertpath++) {
+		X509 *intercert = load_cert(*intercertpath, &ierror);
+		if (intercert == NULL) {
+			g_propagate_error(error, ierror);
+			goto out;
+		}
+
+		if (!CMS_add0_cert(cms, intercert)) {
+			g_set_error(
+					error,
+					R_SIGNATURE_ERROR,
+					R_SIGNATURE_ERROR_CREATE_SIG,
+					"failed to add intermediate certificate: %s", get_openssl_err_string());
+			goto out;
+		}
+	}
+
+	if (!CMS_add1_signer(cms, signcert, pkey, NULL, flags)) {
+		g_set_error(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_CREATE_SIG,
+				"failed to create signature: %s", get_openssl_err_string());
+		goto out;
+	}
+
+	debug_cms_ci(cms);
+
+	if (!i2d_CMS_bio(outsig, cms)) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_SERIALIZE_SIG,
+				"failed to serialize signature");
+		goto out;
+	}
+
+	output_sig = bytes_from_bio(outsig);
+
+	if (!output_sig) {
+		g_set_error_literal(
+				error,
+				R_SIGNATURE_ERROR,
+				R_SIGNATURE_ERROR_UNKNOWN,
+				"Read zero bytes");
+		goto out;
+	}
+
+out:
+	ERR_print_errors_fp(stdout);
+	BIO_free_all(insig);
+	BIO_free_all(outsig);
+	return output_sig;
+}
+
 gchar* get_pubkey_hash(X509 *cert)
 {
 	g_autoptr(GString) string = NULL;
@@ -762,7 +884,7 @@ static gchar* dump_cms(STACK_OF(X509) *x509_certs)
 gchar* sigdata_to_string(GBytes *sig, GError **error)
 {
 	g_autoptr(CMS_ContentInfo) cms = NULL;
-	STACK_OF(X509) *signers = NULL;
+	g_autoptr(R_X509_STACK_POP) signers = NULL;
 	gchar *ret;
 	BIO *insig = bytes_as_bio(sig);
 
@@ -790,7 +912,6 @@ gchar* sigdata_to_string(GBytes *sig, GError **error)
 
 	ret = dump_cms(signers);
 
-	sk_X509_pop_free(signers, X509_free);
 	BIO_free(insig);
 
 	return ret;
@@ -936,7 +1057,7 @@ static gchar* get_cert_time(const ASN1_TIME *time)
 	size = BIO_get_mem_data(mem, &data);
 	ret = g_strndup(data, size);
 
-	BIO_set_close(mem, BIO_CLOSE);
+	g_assert(BIO_set_close(mem, BIO_CLOSE));
 	BIO_free(mem);
 
 	return ret;
@@ -976,25 +1097,38 @@ gchar* format_cert_chain(STACK_OF(X509) *verified_chain)
 	return bio_mem_unwrap(text);
 }
 
-static gchar *cms_get_signers(CMS_ContentInfo *cms, GError **error)
+static STACK_OF(X509) *cms_get_signer_certs(CMS_ContentInfo *cms, GError **error)
 {
-	STACK_OF(X509) *signers = NULL;
-	BIO *text = NULL;
-
 	g_return_val_if_fail(cms != NULL, NULL);
 	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
 
-	signers = CMS_get0_signers(cms);
+	g_autoptr(R_X509_STACK) signers = CMS_get0_signers(cms);
 	if (signers == NULL) {
 		g_set_error_literal(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_GET_SIGNER,
 				"Failed to obtain signer info");
-		goto out;
+		return NULL;
 	}
 
-	text = BIO_new(BIO_s_mem());
+	return g_steal_pointer(&signers);
+}
+
+static gchar *cms_get_signers_string(CMS_ContentInfo *cms, GError **error)
+{
+	GError *ierror = NULL;
+
+	g_return_val_if_fail(cms != NULL, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	g_autoptr(R_X509_STACK) signers = cms_get_signer_certs(cms, &ierror);
+	if (signers == NULL) {
+		g_propagate_error(error, ierror);
+		return NULL;
+	}
+
+	BIO *text = BIO_new(BIO_s_mem());
 	BIO_printf(text, "'");
 	for (int i = 0; i < sk_X509_num(signers); i++) {
 		if (i)
@@ -1003,58 +1137,87 @@ static gchar *cms_get_signers(CMS_ContentInfo *cms, GError **error)
 	}
 	BIO_printf(text, "'");
 
-out:
-	if (signers)
-		sk_X509_free(signers);
-	if (text)
-		return bio_mem_unwrap(text);
-	else
-		return NULL;
+	return bio_mem_unwrap(text);
+}
+
+static gboolean cms_check_signer_cns(CMS_ContentInfo *cms, GError **error)
+{
+	GError *ierror = NULL;
+
+	g_return_val_if_fail(cms != NULL, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	const GStrv allowed_cns = r_context()->config->keyring_allowed_signer_cns;
+
+	// config option not used, so the cert is valid for all devices and we can skip everything
+	if (!allowed_cns) {
+		return TRUE;
+	}
+
+	g_autoptr(R_X509_STACK) signers = cms_get_signer_certs(cms, &ierror);
+	if (signers == NULL) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
+	for (int i = 0; i < sk_X509_num(signers); i++) {
+		X509_NAME *current_signer = X509_get_subject_name(sk_X509_value(signers, i));
+		int index = X509_NAME_get_index_by_NID(current_signer, NID_commonName, -1);
+		// no matching CN entry in the current signer found, so check next signer
+		if (index == -1)
+			continue;
+
+		const X509_NAME_ENTRY *cn = X509_NAME_get_entry(current_signer, index);
+		// as soon as one matching entry is found, device is eligible to use this update
+		const unsigned char* cn_value = ASN1_STRING_get0_data(X509_NAME_ENTRY_get_data(cn));
+		if (g_strv_contains((const gchar *const *)allowed_cns, (gchar*)cn_value)) {
+			return TRUE;
+		}
+	}
+
+	g_set_error_literal(
+			error,
+			R_SIGNATURE_ERROR,
+			R_SIGNATURE_ERROR_SIGNER_CN_FORBIDDEN,
+			"No signature CN matching allowed-signer-cns found");
+	return FALSE;
 }
 
 gboolean cms_get_cert_chain(CMS_ContentInfo *cms, X509_STORE *store, STACK_OF(X509) **verified_chain, GError **error)
 {
-	STACK_OF(X509) *signers = NULL;
-	STACK_OF(X509) *intercerts = NULL;
-	X509_STORE_CTX *cert_ctx = NULL;
-	gint signer_cnt;
-	gboolean res = FALSE;
+	GError *ierror = NULL;
 
 	g_return_val_if_fail(cms != NULL, FALSE);
 	g_return_val_if_fail(store != NULL, FALSE);
 	g_return_val_if_fail(verified_chain == NULL || *verified_chain == NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	signers = CMS_get0_signers(cms);
+	g_autoptr(R_X509_STACK) signers = cms_get_signer_certs(cms, &ierror);
 	if (signers == NULL) {
-		g_set_error_literal(
-				error,
-				R_SIGNATURE_ERROR,
-				R_SIGNATURE_ERROR_GET_SIGNER,
-				"Failed to obtain signer info");
-		goto out;
+		g_propagate_error(error, ierror);
+		return FALSE;
 	}
 
-	signer_cnt = sk_X509_num(signers);
+	gint signer_cnt = sk_X509_num(signers);
 	if (signer_cnt != 1) {
 		g_set_error(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_NUM_SIGNER,
 				"Unsupported number of signers: %d", signer_cnt);
-		goto out;
+		return FALSE;
 	}
 
-	intercerts = CMS_get1_certs(cms);
+	g_autoptr(R_X509_STACK_POP) intercerts = CMS_get1_certs(cms);
 
-	cert_ctx = X509_STORE_CTX_new();
+	g_autoptr(X509_STORE_CTX) cert_ctx = X509_STORE_CTX_new();
 	if (cert_ctx == NULL) {
 		g_set_error_literal(
 				error,
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_X509_CTX_NEW,
 				"Failed to allocate new X509 CTX store");
-		goto out;
+		return FALSE;
 	}
 
 	if (!X509_STORE_CTX_init(cert_ctx, store, sk_X509_value(signers, 0), intercerts)) {
@@ -1063,7 +1226,7 @@ gboolean cms_get_cert_chain(CMS_ContentInfo *cms, X509_STORE *store, STACK_OF(X5
 				R_SIGNATURE_ERROR,
 				R_SIGNATURE_ERROR_X509_CTX_INIT,
 				"Failed to init new X509 CTX store");
-		goto out;
+		return FALSE;
 	}
 
 	if (X509_verify_cert(cert_ctx) != 1) {
@@ -1073,7 +1236,7 @@ gboolean cms_get_cert_chain(CMS_ContentInfo *cms, X509_STORE *store, STACK_OF(X5
 				R_SIGNATURE_ERROR_VERIFY_CERT,
 				"Failed to verify X509 cert: %s",
 				X509_verify_cert_error_string(X509_STORE_CTX_get_error(cert_ctx)));
-		goto out;
+		return FALSE;
 	}
 
 	*verified_chain = X509_STORE_CTX_get1_chain(cert_ctx);
@@ -1083,16 +1246,7 @@ gboolean cms_get_cert_chain(CMS_ContentInfo *cms, X509_STORE *store, STACK_OF(X5
 
 	g_debug("Got %d elements for trust chain", sk_X509_num(*verified_chain));
 
-	res = TRUE;
-out:
-	if (cert_ctx)
-		X509_STORE_CTX_free(cert_ctx);
-	if (intercerts)
-		sk_X509_pop_free(intercerts, X509_free);
-	if (signers)
-		sk_X509_free(signers);
-
-	return res;
+	return TRUE;
 }
 
 /* while OpenSSL 1.1.x provides a function for converting ASN1_TIME to tm,
@@ -1106,7 +1260,7 @@ static gboolean asn1_time_to_tm(const ASN1_TIME *intime, struct tm *tm)
 	BIO *mem;
 	long size;
 	gchar *data;
-	g_autofree gchar *ret;
+	g_autofree gchar *ret = NULL;
 
 	mem = BIO_new(BIO_s_mem());
 
@@ -1120,7 +1274,7 @@ static gboolean asn1_time_to_tm(const ASN1_TIME *intime, struct tm *tm)
 	if (!strptime(ret, "%b %d %H:%M:%S %Y GMT", tm))
 		return FALSE;
 
-	BIO_set_close(mem, BIO_CLOSE);
+	g_assert(BIO_set_close(mem, BIO_CLOSE));
 	BIO_free(mem);
 
 	return TRUE;
@@ -1272,10 +1426,11 @@ gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_C
 	BIO *incontent = NULL;
 	BIO *insig = bytes_as_bio(sig);
 	BIO *outcontent = BIO_new(BIO_s_mem());
-	g_autofree gchar *signers = NULL;
+	g_autofree gchar *signers_string = NULL;
 	gboolean res = FALSE;
 	gboolean verified = FALSE;
 	gboolean detached;
+	unsigned int verify_flags = CMS_BINARY;
 
 	g_return_val_if_fail(sig != NULL, FALSE);
 	g_return_val_if_fail(store != NULL, FALSE);
@@ -1370,9 +1525,9 @@ gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_C
 	}
 
 	if (detached)
-		verified = CMS_verify(icms, NULL, store, incontent, NULL, CMS_DETACHED | CMS_BINARY);
+		verified = CMS_verify(icms, NULL, store, incontent, NULL, verify_flags | CMS_DETACHED);
 	else
-		verified = CMS_verify(icms, NULL, store, NULL, outcontent, CMS_BINARY);
+		verified = CMS_verify(icms, NULL, store, NULL, outcontent, verify_flags);
 	if (!verified) {
 		g_set_error(
 				error,
@@ -1382,12 +1537,17 @@ gboolean cms_verify_bytes(GBytes *content, GBytes *sig, X509_STORE *store, CMS_C
 		goto out;
 	}
 
-	signers = cms_get_signers(icms, &ierror);
-	if (!signers) {
+	if (!cms_check_signer_cns(icms, &ierror)) {
 		g_propagate_error(error, ierror);
 		goto out;
 	}
-	g_message("Verified %s signature by %s", detached ? "detached" : "inline", signers);
+
+	signers_string = cms_get_signers_string(icms, &ierror);
+	if (!signers_string) {
+		g_propagate_error(error, ierror);
+		goto out;
+	}
+	g_message("Verified %s signature by %s", detached ? "detached" : "inline", signers_string);
 
 	if (!detached) {
 		GBytes *tmp = bytes_from_bio(outcontent);
@@ -1550,7 +1710,7 @@ GBytes *cms_encrypt(GBytes *content, gchar **recipients, GError **error)
 	GError *ierror = NULL;
 	BIO *incontent = NULL;
 	BIO *outsig = BIO_new(BIO_s_mem());
-	STACK_OF(X509) *recipcerts = NULL;
+	g_autoptr(R_X509_STACK_POP) recipcerts = NULL;
 	g_autoptr(CMS_ContentInfo) cms = NULL;
 	GBytes *res = NULL;
 
@@ -1564,7 +1724,7 @@ GBytes *cms_encrypt(GBytes *content, gchar **recipients, GError **error)
 
 	/* load all recipient certificates from all provided PEM files */
 	for (gchar **recipcertpath = recipients; recipcertpath && *recipcertpath != NULL; recipcertpath++) {
-		STACK_OF(X509) *filecerts = load_certs_from_file(*recipcertpath, &ierror);
+		g_autoptr(R_X509_STACK) filecerts = load_certs_from_file(*recipcertpath, &ierror);
 		if (filecerts == NULL) {
 			g_propagate_error(error, ierror);
 			goto out;
@@ -1574,8 +1734,6 @@ GBytes *cms_encrypt(GBytes *content, gchar **recipients, GError **error)
 		for (gint i = 0; i < sk_X509_num(filecerts); i++) {
 			sk_X509_push(recipcerts, sk_X509_value(filecerts, i));
 		}
-
-		sk_X509_free(filecerts);
 	}
 
 	cms = CMS_encrypt(recipcerts, incontent, EVP_aes_256_cbc(), CMS_BINARY);
@@ -1610,7 +1768,6 @@ GBytes *cms_encrypt(GBytes *content, gchar **recipients, GError **error)
 
 out:
 	ERR_print_errors_fp(stdout);
-	sk_X509_pop_free(recipcerts, X509_free);
 	BIO_free_all(incontent);
 	BIO_free_all(outsig);
 	return res;
