@@ -32,14 +32,39 @@ GQuark r_update_error_quark(void)
 	return g_quark_from_static_string("r_update_error_quark");
 }
 
-static gboolean check_image_size(int fd, const RaucImage *image, GError **error)
+/**
+ * Checks if given image fits into the (block) device.
+ *
+ * Checks if the image size fits into the block device referred to by the
+ * provided file descriptor.
+ * If the size cannot be determined, the image is assumed to fit.
+ *
+ * Additionally, if a slot is given and the slot has a 'size-limit' set, the
+ * image is checked against this limit first.
+ *
+ * @param fd file descriptor of device to check
+ * @param slot slot (e.g. for optional size-limit check)
+ * @param image image to check
+ * @param error return location for a GError, or NULL
+ *
+ * @return TRUE if image fits into device (or size-limit) or if no size can be determined. FALSE otherwise.
+ */
+static gboolean check_image_size(int fd, const RaucSlot *slot, const RaucImage *image, GError **error)
 {
 	GError *ierror = NULL;
 	goffset dev_size;
 
+	g_return_val_if_fail(slot, FALSE);
 	g_return_val_if_fail(image, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
+	if (slot->size_limit > 0 && (guint64)image->checksum.size > slot->size_limit) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
+				"Image size (%"G_GOFFSET_FORMAT " bytes) is larger than size-limit (%"G_GUINT64_FORMAT " bytes).",
+				image->checksum.size, slot->size_limit);
+		return FALSE;
+	}
+
 	dev_size = get_device_size(fd, &ierror);
 	if (g_error_matches(ierror, R_UTILS_ERROR, R_UTILS_ERROR_INAPPROPRIATE_IOCTL)) {
 		g_clear_error(&ierror);
@@ -47,6 +72,13 @@ static gboolean check_image_size(int fd, const RaucImage *image, GError **error)
 		return TRUE;
 	}
 
+	/* Sanity check for size limit being < device size.
+	 * Should be moved to a proper location for configuration vs. target checks (maybe context setup) and become an error later. */
+	if (slot->size_limit > 0 && slot->size_limit > (guint64)dev_size) {
+		g_warning("The size-limit (%"G_GUINT64_FORMAT " bytes) exceeds actual device size (%"G_GOFFSET_FORMAT " bytes).",
+				slot->size_limit, dev_size);
+	}
+
 	if (dev_size < image->checksum.size) {
 		if (ierror) {
 			g_propagate_error(error, ierror);
@@ -68,6 +100,7 @@ static gboolean clear_slot(RaucSlot *slot, GError **error)
 	static gchar zerobuf[CLEAR_BLOCK_SIZE] = {};
 	g_autoptr(GOutputStream) outstream = NULL;
 	gint write_count = 0;
+	guint64 total_written = 0;
 
 	outstream = G_OUTPUT_STREAM(r_unix_output_stream_open_device(slot->device, NULL, &ierror));
 	if (outstream == NULL) {
@@ -76,20 +109,39 @@ static gboolean clear_slot(RaucSlot *slot, GError **error)
 	}
 
 	while (write_count != -1) {
-		write_count = g_output_stream_write(outstream, zerobuf, CLEAR_BLOCK_SIZE, NULL,
+		/* cap writes to slot->size_limit if set */
+		gsize write_size = CLEAR_BLOCK_SIZE;
+		if (slot->size_limit > 0 &&
+		    total_written + write_size > slot->size_limit)
+			write_size = slot->size_limit - total_written;
+
+		write_count = g_output_stream_write(outstream, zerobuf, write_size, NULL,
 				&ierror);
 		/*
 		 * G_IO_ERROR_NO_SPACE is expected here, because the block
 		 * device is cleared completely
 		 */
-		if (write_count == -1 &&
-		    !g_error_matches(ierror, G_IO_ERROR, G_IO_ERROR_NO_SPACE)) {
-			g_propagate_prefixed_error(error, ierror,
-					"failed clearing block device: ");
-			return FALSE;
+		if (write_count == -1) {
+			if (g_error_matches(ierror, G_IO_ERROR, G_IO_ERROR_NO_SPACE)) {
+				g_clear_error(&ierror);
+				break;
+			} else {
+				g_propagate_prefixed_error(error, ierror,
+						"failed clearing block device: ");
+				return FALSE;
+			}
 		}
+
+		total_written += write_count;
+		if (slot->size_limit > 0 && total_written >= slot->size_limit)
+			break;
 	}
 
+	if (slot->size_limit > 0)
+		g_message("Cleared first %"G_GOFFSET_FORMAT " bytes on %s", total_written, slot->device);
+	else
+		g_debug("Cleared %"G_GOFFSET_FORMAT " bytes on %s", total_written, slot->device);
+
 	if (!g_output_stream_close(outstream, NULL, &ierror)) {
 		g_propagate_error(error, ierror);
 		return FALSE;
@@ -189,7 +241,6 @@ static gboolean splice_with_progress(GUnixInputStream *image_stream,
 	struct stat stat = {};
 	ssize_t out_size = 0;
 	goffset sum_size = 0;
-	gint last_percent = -1, percent;
 
 	if (fstat(in_fd, &stat)) {
 		int err = errno;
@@ -212,12 +263,9 @@ static gboolean splice_with_progress(GUnixInputStream *image_stream,
 
 		sum_size += out_size;
 
-		percent = sum_size * 100 / stat.st_size;
 		/* emit progress info (but only when in progress context) */
-		if (r_context()->progress && percent != last_percent) {
-			last_percent = percent;
-			r_context_set_step_percentage("copy_image", percent);
-		}
+		if (r_context()->progress)
+			r_context_set_step_percentage("copy_image", sum_size * 100 / stat.st_size);
 	} while (out_size);
 
 	return TRUE;
@@ -376,8 +424,7 @@ static gboolean write_boot_switch_partition(RaucImage *image, const gchar *devic
 		GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
-	int out_fd = -1;
+	g_auto(filedesc) out_fd = -1;
 	g_autoptr(GUnixOutputStream) outstream = NULL;
 
 	g_return_val_if_fail(image, FALSE);
@@ -390,8 +437,7 @@ static gboolean write_boot_switch_partition(RaucImage *image, const gchar *devic
 		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
 				"Opening output device failed: %s",
 				strerror(errno));
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
 	if (lseek(out_fd, dest_partition->start, SEEK_SET) !=
@@ -399,35 +445,27 @@ static gboolean write_boot_switch_partition(RaucImage *image, const gchar *devic
 		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
 				"Failed to set file to position %"G_GUINT64_FORMAT,
 				dest_partition->start);
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
 	outstream = G_UNIX_OUTPUT_STREAM(g_unix_output_stream_new(out_fd, FALSE));
 	if (outstream == NULL) {
 		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
 				"Failed to create output stream");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
-	res = copy_raw_image(image, outstream, len_header_last, &ierror);
-	if (!res) {
+	if (!copy_raw_image(image, outstream, len_header_last, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-	res = g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror);
-	if (!res) {
+	if (!g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-out:
-	if (out_fd >= 0)
-		close(out_fd);
-
-	return res;
+	return TRUE;
 }
 
 static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const gchar *seed, const gchar *store, const gchar *tmpdir, GError **error)
@@ -436,7 +474,6 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 	g_autoptr(GSubprocess) sproc = NULL;
 	g_auto(GStrv) casync_argvp = NULL;
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 	g_autoptr(GPtrArray) args = g_ptr_array_new_full(5, g_free);
 
 	if (r_context()->config->use_desync)
@@ -458,15 +495,12 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 		g_ptr_array_add(args, g_strdup("--seed-output=no"));
 
 	if (r_context()->config->casync_install_args != NULL) {
-		gboolean parse_res = FALSE;
-		parse_res = g_shell_parse_argv(r_context()->config->casync_install_args, NULL, &casync_argvp, &ierror);
-		if (!parse_res) {
-			res = parse_res;
+		if (!g_shell_parse_argv(r_context()->config->casync_install_args, NULL, &casync_argvp, &ierror)) {
 			g_propagate_prefixed_error(
 					error,
 					ierror,
 					"Failed to parse casync extra args: ");
-			goto out;
+			return FALSE;
 		}
 		r_ptr_array_addv(args, casync_argvp, TRUE);
 	}
@@ -483,8 +517,7 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 			g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
 					"Failed to dup() output file descriptor: %s",
 					strerror(errno));
-			res = FALSE;
-			goto out;
+			return FALSE;
 		}
 		g_subprocess_launcher_take_stdout_fd(launcher, out_fd_copy);
 	}
@@ -502,20 +535,18 @@ static gboolean casync_extract(RaucImage *image, gchar *dest, int out_fd, const
 				error,
 				ierror,
 				"failed to start casync extract: ");
-		goto out;
+		return FALSE;
 	}
 
-	res = g_subprocess_wait_check(sproc, NULL, &ierror);
-	if (!res) {
+	if (!g_subprocess_wait_check(sproc, NULL, &ierror)) {
 		g_propagate_prefixed_error(
 				error,
 				ierror,
 				"failed to run casync extract: ");
-		goto out;
+		return FALSE;
 	}
 
-out:
-	return res;
+	return TRUE;
 }
 
 static RaucSlot *get_active_slot_class_member(gchar *slotclass)
@@ -641,6 +672,20 @@ unmount_out:
 	return res;
 }
 
+/**
+ * Writes given RaucImage to the device referred to by the given RaucSlot.
+ *
+ * Checks if the provided RaucImage fits into the target device before the copy
+ * process starts.
+ *
+ * Copying is done using simple dd-like raw data copying.
+ *
+ * @param image RaucImage to write
+ * @param slot RaucSlot to copy to
+ * @param error return location for a GError, or NULL
+ *
+ * @return TRUE on success. FALSE on error.
+ */
 static gboolean copy_raw_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
 {
 	g_autoptr(GUnixOutputStream) outstream = NULL;
@@ -657,7 +702,7 @@ static gboolean copy_raw_image_to_dev(RaucImage *image, RaucSlot *slot, GError *
 	}
 
 	/* check size */
-	if (!check_image_size(g_unix_output_stream_get_fd(outstream), image, &ierror)) {
+	if (!check_image_size(g_unix_output_stream_get_fd(outstream), slot, image, &ierror)) {
 		res = FALSE;
 		g_propagate_error(error, ierror);
 		goto out;
@@ -711,12 +756,18 @@ static gboolean copy_block_hash_index_image_to_dev(RaucImage *image, RaucSlot *s
 		res = FALSE;
 		goto out;
 	}
-	if (!check_image_size(tmp->data_fd, image, &ierror)) {
+	if (!check_image_size(tmp->data_fd, slot, image, &ierror)) {
 		g_propagate_error(error, ierror);
 		res = FALSE;
 		goto out;
 	}
 
+	/* emit progress info (when in progress context) since a new
+	 * target_slot hash index might have been generated and that could
+	 * have taken some time. */
+	if (r_context()->progress)
+		r_context_set_step_percentage("copy_image", R_HASH_INDEX_GEN_PROGRESS_SPAN);
+
 	g_ptr_array_add(sources, g_steal_pointer(&tmp));
 
 	/* Open and append seed slot. */
@@ -733,6 +784,12 @@ static gboolean copy_block_hash_index_image_to_dev(RaucImage *image, RaucSlot *s
 		g_message("No active slot available to use as seed for %s", image->slotclass);
 	}
 
+	/* emit progress info (when in progress context) since a new
+	 * active_slot hash index might have been generated and that could
+	 * have taken some time. */
+	if (r_context()->progress)
+		r_context_set_step_percentage("copy_image", R_HASH_INDEX_GEN_PROGRESS_SPAN * 2);
+
 	/* Open and append source image. */
 	tmp = r_hash_index_open_image("source_image", image, &ierror);
 	if (!tmp) {
@@ -845,6 +902,12 @@ static gboolean copy_block_hash_index_image_to_dev(RaucImage *image, RaucSlot *s
 			target_written->invalid_from = c+1;
 			target_old->invalid_below = c;
 		}
+
+		/* emit progress info (but only when in progress context).
+		 * Since the first 10 percent are reserved for the hash index
+		 * generation, we just set the last 90 percent here. */
+		if (r_context()->progress)
+			r_context_set_step_percentage("copy_image", (R_HASH_INDEX_GEN_PROGRESS_SPAN * 2) + (c + 1) * (100 - (R_HASH_INDEX_GEN_PROGRESS_SPAN * 2)) / chunk_count);
 	}
 
 	/* Seek after the written data so this behaves similar to the simpler write helpers */
@@ -908,6 +971,18 @@ static gboolean copy_adaptive_image_to_dev(RaucImage *image, RaucSlot *slot, GEr
 	return FALSE;
 }
 
+/**
+ * Writes given RaucImage to the device referred to by the given RaucSlot.
+ *
+ * Handles both casync .caibx images and adaptive methods, if present.
+ * Otherwise (or on adaptive errors), it performs a raw copy.
+ *
+ * @param image RaucImage to write
+ * @param slot RaucSlot to copy to
+ * @param error return location for a GError, or NULL
+ *
+ * @return TRUE on success. FALSE on error.
+ */
 static gboolean write_image_to_dev(RaucImage *image, RaucSlot *slot, GError **error)
 {
 	GError *ierror = NULL;
@@ -936,6 +1011,10 @@ static gboolean write_image_to_dev(RaucImage *image, RaucSlot *slot, GError **er
 				g_info("%s", ierror->message);
 			} else {
 				g_warning("Continuing after adaptive mode error: %s", ierror->message);
+				/* Note that step progress can already be at a certain percentage at this point
+				 * and will be reset to 0 when normal copying runs instead.
+				 * This isn't propagated but high level progress may stall until step progress
+				 * exceeds the current value again. */
 			}
 			g_clear_error(&ierror);
 			/* Continue with full copy */
@@ -1314,9 +1393,12 @@ static gboolean run_slot_hook_extra_env(const gchar *hook_name, const gchar *hoo
 	gboolean res = FALSE;
 	RaucBundle *bundle;
 
-	g_assert_nonnull(slot);
-	g_assert_nonnull(slot->name);
-	g_assert_nonnull(slot->sclass);
+	g_return_val_if_fail(hook_name, FALSE);
+	g_return_val_if_fail(hook_cmd, FALSE);
+	g_return_val_if_fail(slot, FALSE);
+	g_return_val_if_fail(slot->name, FALSE);
+	g_return_val_if_fail(slot->sclass, FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
 	g_message("Running slot hook '%s' for %s", hook_cmd, slot->name);
 
@@ -1420,9 +1502,10 @@ static gboolean mount_and_run_slot_hook(const gchar *hook_name, const gchar *hoo
 	GError *ierror = NULL;
 	gboolean res = FALSE;
 
+	g_return_val_if_fail(hook_name, FALSE);
+	g_return_val_if_fail(hook_cmd, FALSE);
+	g_return_val_if_fail(slot, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
-	g_assert_nonnull(hook_name);
-	g_assert_nonnull(hook_cmd);
 
 	/* mount slot */
 	g_message("Mounting slot %s", slot->device);
@@ -1763,12 +1846,15 @@ static gboolean archive_to_ext4_handler(RaucImage *image, RaucSlot *dest_slot, c
 		goto out;
 	}
 
-	/* extract tar into mounted ext4 volume */
-	g_message("Extracting %s to %s", image->filename, dest_slot->mount_point);
-	res = unpack_archive(image, dest_slot->mount_point, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto unmount_out;
+	/* 'emptyfs' image type does not have an archive included. So skip extraction. */
+	if (g_strcmp0(image->type, "emptyfs") != 0) {
+		/* extract tar into mounted ext4 volume */
+		g_message("Extracting %s to %s", image->filename, dest_slot->mount_point);
+		res = unpack_archive(image, dest_slot->mount_point, &ierror);
+		if (!res) {
+			g_propagate_error(error, ierror);
+			goto unmount_out;
+		}
 	}
 
 	/* run slot post install hook if enabled */
@@ -1870,130 +1956,112 @@ out:
 static gboolean img_to_nor_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	/* run slot pre install hook if enabled */
 	if (hook_name && image->hooks.pre_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	/* erase */
 	g_message("erasing slot device %s", dest_slot->device);
-	res = flash_format_slot(dest_slot->device, &ierror);
-	if (!res) {
+	if (!flash_format_slot(dest_slot->device, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* write */
 	g_message("writing slot device %s", dest_slot->device);
-	res = nor_write_slot(image->filename, dest_slot->device, &ierror);
-	if (!res) {
+	if (!nor_write_slot(image->filename, dest_slot->device, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* run slot post install hook if enabled */
 	if (hook_name && image->hooks.post_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean img_to_nand_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	/* run slot pre install hook if enabled */
 	if (hook_name && image->hooks.pre_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	/* erase */
 	g_message("erasing slot device %s", dest_slot->device);
-	res = flash_format_slot(dest_slot->device, &ierror);
-	if (!res) {
+	if (!flash_format_slot(dest_slot->device, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* write */
 	g_message("writing slot device %s", dest_slot->device);
-	res = nand_write_slot(image->filename, dest_slot->device, &ierror);
-	if (!res) {
+	if (!nand_write_slot(image->filename, dest_slot->device, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* run slot post install hook if enabled */
 	if (hook_name && image->hooks.post_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean img_to_fs_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	/* run slot pre install hook if enabled */
 	if (hook_name && image->hooks.pre_install) {
-		res = mount_and_run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!mount_and_run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	/* copy */
-	res = write_image_to_dev(image, dest_slot, &ierror);
-	if (!res) {
+	if (!write_image_to_dev(image, dest_slot, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	if (dest_slot->resize && g_strcmp0(dest_slot->type, "ext4") == 0) {
 		g_message("Resizing %s", dest_slot->device);
-		res = ext4_resize_slot(dest_slot, &ierror);
-		if (!res) {
+		if (!ext4_resize_slot(dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	/* run slot post install hook if enabled */
 	if (hook_name && image->hooks.post_install) {
-		res = mount_and_run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!mount_and_run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean img_to_boot_mbr_switch_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
@@ -2180,64 +2248,29 @@ out:
 #endif
 
 #if ENABLE_EMMC_BOOT_SUPPORT == 1
-static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
+/**
+ * Copies given image to eMMC boot partition.
+ *
+ * @param image Image to copy
+ * @param slot Slot to copy to
+ * @param hook_name name of the hook, or NULL
+ * @param vars additional env vars to pass, or NULL
+ * @param error return location for a GError, or NULL
+ */
+static gboolean copy_img_to_emmc_bootpart(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GHashTable *vars, GError **error)
 {
 	gboolean res = FALSE;
-	int out_fd;
-	gint part_active;
-	g_autofree gchar *part_active_str = NULL;
-	g_autofree gchar *realdev = NULL;
-	gint part_active_after;
-	g_autoptr(GUnixOutputStream) outstream = NULL;
 	GError *ierror = NULL;
-	g_autoptr(RaucSlot) part_slot = NULL;
-	g_autoptr(GHashTable) vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
-
-	realdev = r_realpath(dest_slot->device);
-	if (!realdev) {
-		g_set_error(error,
-				R_UPDATE_ERROR,
-				R_UPDATE_ERROR_FAILED,
-				"Can't resolve eMMC device %s", dest_slot->device);
-		goto out;
-	}
-
-	/* read active boot partition from ext_csd */
-	res = r_emmc_read_bootpart(realdev, &part_active, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto out;
-	}
-
-	if (part_active == -1) {
-		g_warning("eMMC device was not enabled for booting, yet. Ignoring.");
-		/* For simplicity: Consider boot1 to be active */
-		part_active = 1;
-	} else {
-		g_message("Found active eMMC boot partition %sboot%d", realdev, part_active);
-	}
-
-	/* create a temporary RaucSlot with the actual (currently inactive) boot
-	 * partition as device.
-	 */
-	part_slot = g_new0(RaucSlot, 1);
-	part_slot->device = g_strdup_printf(
-			"%sboot%d",
-			realdev,
-			INACTIVE_BOOT_PARTITION(part_active));
 
 	/* disable read-only on determined eMMC boot partition */
 	g_debug("Disabling read-only mode of slot device partition %s",
-			part_slot->device);
-	res = r_emmc_force_part_rw(part_slot->device, &ierror);
+			dest_slot->device);
+	res = r_emmc_force_part_rw(dest_slot->device, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
 	}
 
-	g_hash_table_insert(vars, g_strdup("RAUC_BOOT_PARTITION_ACTIVATING"),
-			g_strdup_printf("%d", INACTIVE_BOOT_PARTITION(part_active)));
-
 	/* run slot pre install hook if enabled */
 	if (hook_name && image->hooks.pre_install) {
 		res = run_slot_hook_extra_env(
@@ -2254,37 +2287,19 @@ static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot,
 	}
 
 	/* clear block device partition */
-	g_message("Clearing slot device %s", part_slot->device);
-	res = clear_slot(part_slot, &ierror);
+	g_message("Clearing slot device %s", dest_slot->device);
+	res = clear_slot(dest_slot, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
 	}
 
-	/* open */
-	g_message("Opening slot device partition %s", part_slot->device);
-	outstream = r_unix_output_stream_open_device(part_slot->device, &out_fd, &ierror);
-	if (outstream == NULL) {
+	if (!copy_raw_image_to_dev(image, dest_slot, &ierror)) {
 		g_propagate_error(error, ierror);
 		res = FALSE;
 		goto out;
 	}
 
-	/* copy */
-	g_message("Copying image to slot device partition %s",
-			part_slot->device);
-	res = copy_raw_image(image, outstream, 0, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto out;
-	}
-
-	res = g_output_stream_close(G_OUTPUT_STREAM(outstream), NULL, &ierror);
-	if (!res) {
-		g_propagate_error(error, ierror);
-		goto out;
-	}
-
 	/* run slot post install hook if enabled */
 	if (hook_name && image->hooks.post_install) {
 		res = run_slot_hook_extra_env(
@@ -2302,19 +2317,85 @@ static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot,
 
 	/* re-enable read-only on determined eMMC boot partition */
 	g_debug("Reenabling read-only mode of slot device partition %s",
-			part_slot->device);
-	res = r_emmc_force_part_ro(part_slot->device, &ierror);
+			dest_slot->device);
+	res = r_emmc_force_part_ro(dest_slot->device, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
 		goto out;
 	}
 
+out:
+	if (!res)
+		r_emmc_force_part_ro(dest_slot->device, NULL);
+
+	return res;
+}
+
+static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
+{
+	gboolean res = FALSE;
+	gint part_active;
+	g_autofree gchar *realdev = NULL;
+	gint part_active_after;
+	GError *ierror = NULL;
+	g_autoptr(RaucSlot) part_slot = NULL;
+	g_autoptr(GHashTable) vars = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+
+	realdev = r_realpath(dest_slot->device);
+	if (!realdev) {
+		g_set_error(error,
+				R_UPDATE_ERROR,
+				R_UPDATE_ERROR_FAILED,
+				"Can't resolve eMMC device %s", dest_slot->device);
+		return FALSE;
+	}
+
+	/* read active boot partition from ext_csd */
+	res = r_emmc_read_bootpart(realdev, &part_active, &ierror);
+	if (!res) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
+	if (part_active == -1) {
+		g_warning("eMMC device was not enabled for booting, yet. Ignoring.");
+		/* For simplicity: Consider boot1 to be active */
+		part_active = 1;
+	} else {
+		g_message("Found active eMMC boot partition %sboot%d", realdev, part_active);
+	}
+
+	/* create a temporary RaucSlot with the actual (currently inactive) boot
+	 * partition as device.
+	 */
+	part_slot = g_new0(RaucSlot, 1);
+	part_slot->device = g_strdup_printf(
+			"%sboot%d",
+			realdev,
+			INACTIVE_BOOT_PARTITION(part_active));
+	part_slot->size_limit = dest_slot->size_limit;
+
+	g_hash_table_insert(vars, g_strdup("RAUC_BOOT_PARTITION_ACTIVATING"),
+			g_strdup_printf("%d", INACTIVE_BOOT_PARTITION(part_active)));
+	g_hash_table_insert(vars, g_strdup("RAUC_BOOT_SIZE_LIMIT"),
+			g_strdup_printf("%"G_GUINT64_FORMAT, dest_slot->size_limit));
+
+	/* To keep RAUC_SLOT_DEVICE consistent, override with the eMMC base
+	 * device here. */
+	g_hash_table_insert(vars, g_strdup("RAUC_SLOT_DEVICE"), g_strdup(realdev));
+
+	res = copy_img_to_emmc_bootpart(image, part_slot, hook_name, vars, &ierror);
+	if (!res) {
+		g_propagate_error(error, ierror);
+		return FALSE;
+	}
+
 	/* toggle active boot partition in ext_csd register; do this explicitly on
 	 * determined boot partition to force the kernel to switch to the partition;
 	 * for simplicity reasons: in case the user partition is active use
 	 * mmcblkXboot1, in case no partition is active use mmcblkXboot0
 	 */
-	g_debug("Toggling active eMMC boot partition %s -> %s", part_active_str,
+	g_debug("Toggling active eMMC boot partition %sboot%d -> %s", realdev, part_active,
 			part_slot->device);
 	res = r_emmc_write_bootpart(
 			part_slot->device,
@@ -2322,7 +2403,7 @@ static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot,
 			&ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* sanity check: read active boot partition from ext_csd
@@ -2333,24 +2414,18 @@ static gboolean img_to_boot_emmc_handler(RaucImage *image, RaucSlot *dest_slot,
 	res = r_emmc_read_bootpart(realdev, &part_active_after, &ierror);
 	if (!res) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	if (part_active == part_active_after) {
 		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED,
 				"Toggling the boot partition failed! Your kernel is most-likely affected by the ioctl ext_csd bug: see https://rauc.readthedocs.io/en/latest/advanced.html#update-bootloader-in-emmc-boot-partitions");
-		res = FALSE;
-		goto out;
+		return FALSE;
 	}
 
 	g_message("Boot partition %s is now active", part_slot->device);
 
-out:
-	/* ensure that the eMMC boot partition is read-only afterwards */
-	if (!res && part_slot)
-		r_emmc_force_part_ro(part_slot->device, NULL);
-
-	return res;
+	return TRUE;
 }
 #endif
 
@@ -2523,131 +2598,187 @@ static gboolean img_to_boot_raw_fallback_handler(RaucImage *image, RaucSlot *des
 static gboolean img_to_raw_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	/* run slot pre install hook if enabled */
 	if (hook_name && image->hooks.pre_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_PRE_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
 	/* copy */
-	res = write_image_to_dev(image, dest_slot, &ierror);
-	if (!res) {
+	if (!write_image_to_dev(image, dest_slot, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
 	/* run slot post install hook if enabled */
 	if (hook_name && image->hooks.post_install) {
-		res = run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror);
-		if (!res) {
+		if (!run_slot_hook(hook_name, R_SLOT_HOOK_POST_INSTALL, image, dest_slot, &ierror)) {
 			g_propagate_error(error, ierror);
-			goto out;
+			return FALSE;
 		}
 	}
 
-out:
-	return res;
+	return TRUE;
 }
 
 static gboolean hook_install_handler(RaucImage *image, RaucSlot *dest_slot, const gchar *hook_name, GError **error)
 {
 	GError *ierror = NULL;
-	gboolean res = FALSE;
 
 	/* run slot install hook */
-	res = run_slot_hook(hook_name, R_SLOT_HOOK_INSTALL, image, dest_slot, &ierror);
-	if (!res) {
+	if (!run_slot_hook(hook_name, R_SLOT_HOOK_INSTALL, image, dest_slot, &ierror)) {
 		g_propagate_error(error, ierror);
-		goto out;
+		return FALSE;
 	}
 
-out:
-	return res;
+	return TRUE;
 }
+typedef struct {
+	const gchar *type;
+	const gchar *slottype;
+	img_to_slot_handler handler;
+} RaucImageTypeMap;
+
+/* Image type to handler mapping */
+static RaucImageTypeMap image_type_map[] = {
+	/* casync - caibx */
+	{"ext4-caibx", "ext4", img_to_fs_handler},
+	{"raw-caibx", "ext4", img_to_raw_handler},
+
+	{"ext4-caibx", "raw", img_to_raw_handler},
+	{"vfat-caibx", "raw", img_to_raw_handler},
+	{"raw-caibx", "raw", img_to_raw_handler},
+	{"raw-caibx", "vfat", img_to_raw_handler},
+	{"raw-caibx", "jffs2", img_to_raw_handler},
+	{"raw-caibx", "nor", img_to_raw_handler},
+	{"raw-caibx", "nand", img_to_raw_handler},
+	{"squashfs-caibx", "raw", img_to_raw_handler},
+
+	{"squashfs-caibx", "ubivol", img_to_ubivol_handler},
+	{"ubifs-caibx", "ubivol", img_to_ubivol_handler},
+	{"raw-caibx", "ubivol", img_to_ubivol_handler},
+	{"ubifs-caibx", "ubifs", img_to_ubifs_handler},
+	{"raw-caibx", "ubifs", img_to_ubifs_handler},
+	/* casync - caidx */
+	{"caidx", "ext4", archive_to_ext4_handler},
+	{"caidx", "ubifs", archive_to_ubifs_handler},
+	{"caidx", "vfat", archive_to_vfat_handler},
+	/* casync - catar */
+	{"catar", "ext4", archive_to_ext4_handler},
+	/* file system */
+	{"ext4", "ext4", img_to_fs_handler},
+	{"ext4", "raw", img_to_raw_handler},
+	{"vfat", "vfat", img_to_fs_handler},
+	{"vfat", "raw", img_to_raw_handler},
+	{"squashfs", "ubivol", img_to_ubivol_handler},
+	{"squashfs", "raw", img_to_raw_handler},
+	{"ubifs", "ubivol", img_to_ubivol_handler},
+	{"ubifs", "ubifs", img_to_ubifs_handler},
+	/* raw */
+	{"raw", "ext4", img_to_fs_handler},
+	{"raw", "nor", img_to_nor_handler},
+	{"raw", "nand", img_to_nand_handler},
+	{"raw", "ubifs", img_to_ubifs_handler},
+	{"raw", "ubivol", img_to_ubivol_handler},
+	{"raw", "vfat", img_to_fs_handler},
+	{"raw", "raw", img_to_raw_handler},
+	{"raw", "jffs2", img_to_raw_handler},
+	/* archive */
+	{"tar", "ext4", archive_to_ext4_handler},
+	{"tar", "ubifs", archive_to_ubifs_handler},
+	{"tar", "vfat", archive_to_vfat_handler},
+	{"tar", "jffs2", archive_to_jffs2_handler},
+	/* boot-* slot types */
+#if ENABLE_EMMC_BOOT_SUPPORT == 1
+	{"raw", "boot-emmc", img_to_boot_emmc_handler},
+#endif
+	{"vfat", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
+	{"raw", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
+#if ENABLE_GPT == 1
+	{"vfat", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
+	{"ext4", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
+	{"raw", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
+#endif
+	{"raw", "boot-raw-fallback", img_to_boot_raw_fallback_handler},
+	{"emptyfs", "ext4", archive_to_ext4_handler},
+	{NULL, NULL, NULL}
+};
 
 typedef struct {
-	const gchar *src;
-	const gchar *dest;
-	img_to_slot_handler handler;
-} RaucUpdatePair;
+	const gchar *fileext;
+	const gchar *type;
+} RaucFileExtTypeMap;
 
-RaucUpdatePair updatepairs[] = {
-	{"*.ext4.caibx", "ext4", img_to_fs_handler},
-	{"*.ext4.caibx", "raw", img_to_raw_handler},
-	{"*.vfat.caibx", "raw", img_to_raw_handler},
-	{"*.ubifs.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.ubifs.caibx", "ubifs", img_to_ubifs_handler},
-	//{"*.img.caibx", "nand", img_to_nand_handler}, /* unsupported */
-	{"*.img.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.img.caibx", "ubifs", img_to_ubifs_handler},
-	{"*.squashfs.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-lz4.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-lzo.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-xz.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-zst.caibx", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs.caibx", "raw", img_to_raw_handler},
-	{"*.squashfs-lz4.caibx", "raw", img_to_raw_handler},
-	{"*.squashfs-lzo.caibx", "raw", img_to_raw_handler},
-	{"*.squashfs-xz.caibx", "raw", img_to_raw_handler},
-	{"*.squashfs-zst.caibx", "raw", img_to_raw_handler},
-	{"*.caidx", "ext4", archive_to_ext4_handler},
-	{"*.caidx", "ubifs", archive_to_ubifs_handler},
-	{"*.caidx", "vfat", archive_to_vfat_handler},
-	{"*.ext4", "ext4", img_to_fs_handler},
-	{"*.ext4", "raw", img_to_raw_handler},
-	{"*.vfat", "raw", img_to_raw_handler},
-	{"*.squashfs", "raw", img_to_raw_handler},
-	{"*.squashfs-lz4", "raw", img_to_raw_handler},
-	{"*.squashfs-lzo", "raw", img_to_raw_handler},
-	{"*.squashfs-xz", "raw", img_to_raw_handler},
-	{"*.squashfs-zst", "raw", img_to_raw_handler},
-	{"*.vfat", "vfat", img_to_fs_handler},
-	{"*.tar*", "ext4", archive_to_ext4_handler},
-	{"*.catar", "ext4", archive_to_ext4_handler},
-	{"*.tar*", "ubifs", archive_to_ubifs_handler},
-	{"*.tar*", "jffs2", archive_to_jffs2_handler},
-	{"*.tar*", "vfat", archive_to_vfat_handler},
-	{"*.ubifs", "ubivol", img_to_ubivol_handler},
-	{"*.ubifs", "ubifs", img_to_ubifs_handler},
-	{"*.img", "ext4", img_to_fs_handler},
-	{"*.img", "nor", img_to_nor_handler},
-	{"*.img", "nand", img_to_nand_handler},
-	{"*.img", "ubifs", img_to_ubifs_handler},
-	{"*.img", "ubivol", img_to_ubivol_handler},
-	{"*.img", "vfat", img_to_fs_handler},
-	{"*.squashfs", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-lz4", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-lzo", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-xz", "ubivol", img_to_ubivol_handler},
-	{"*.squashfs-zst", "ubivol", img_to_ubivol_handler},
-#if ENABLE_EMMC_BOOT_SUPPORT == 1
-	{"*.img", "boot-emmc", img_to_boot_emmc_handler},
-#endif
-	{"*", "boot-emmc", NULL},
-	{"*.vfat", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
-	{"*.img", "boot-mbr-switch", img_to_boot_mbr_switch_handler},
-	{"*", "boot-mbr-switch", NULL},
-#if ENABLE_GPT == 1
-	{"*.vfat", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
-	{"*.ext4", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
-	{"*.img", "boot-gpt-switch", img_to_boot_gpt_switch_handler},
-#endif
-	{"*", "boot-gpt-switch", NULL},
-	{"*.img", "boot-raw-fallback", img_to_boot_raw_fallback_handler},
-	{"*", "boot-raw-fallback", NULL},
-	{"*.img.caibx", "*", img_to_raw_handler}, /* fallback */
-	{"*.img", "*", img_to_raw_handler}, /* fallback */
-	{0}
+/* For compatibility reasons, this mapping is needed to map the former file
+ * extension-based approach to the image types.
+ * It must be ordered with more specific patterns first (e.g.
+ * '*.squashfs-*.caibx' before '*.squashfs-*') to avoid selecting the wrong
+ * image type. */
+static RaucFileExtTypeMap ext_type_map[] = {
+	{"*.ext4.caibx", "ext4-caibx"},
+	{"*.vfat.caibx", "vfat-caibx"},
+	{"*.ubifs.caibx", "ubifs-caibx"},
+	{"*.img.caibx", "raw-caibx"},
+	{"*.squashfs.caibx", "squashfs-caibx"},
+	{"*.squashfs-*.caibx", "squashfs-caibx"},
+	{"*.catar", "catar"},
+	{"*.caidx", "caidx"},
+	{"*.tar*", "tar"},
+	{"*.tgz", "tar"},
+	{"*.ext4", "ext4"},
+	{"*.vfat", "vfat"},
+	{"*.img", "raw"},
+	{"*.squashfs-*", "squashfs"},
+	{"*.squashfs", "squashfs"},
+	{"*.ubifs", "ubifs"},
 };
 
+const gchar* derive_image_type_from_filename_pattern(const gchar *filename)
+{
+	g_return_val_if_fail(filename, NULL);
+
+	for (unsigned long i = 0; i < G_N_ELEMENTS(ext_type_map); i++) {
+		if (g_pattern_match_simple(ext_type_map[i].fileext, filename)) {
+			return ext_type_map[i].type;
+		}
+	}
+
+	return NULL;
+}
+
+static img_to_slot_handler get_handler_from_type(const gchar *image_type, const gchar *slot_type)
+{
+	g_return_val_if_fail(image_type, NULL);
+	g_return_val_if_fail(slot_type, NULL);
+
+	for (RaucImageTypeMap *map = image_type_map; map->type != NULL; map++) {
+		if (g_strcmp0(map->type, image_type) == 0 &&
+		    g_strcmp0(map->slottype, slot_type) == 0) {
+			return map->handler;
+		}
+	}
+
+	return NULL;
+}
+
+gboolean is_image_type_supported(const gchar *type)
+{
+	g_return_val_if_fail(type, FALSE);
+
+	for (RaucImageTypeMap *map = image_type_map; map->type != NULL; map++) {
+		if (g_strcmp0(map->type, type) == 0) {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 img_to_slot_handler get_update_handler(RaucImage *mfimage, RaucSlot *dest_slot, GError **error)
 {
-	const gchar *src = mfimage->filename;
 	const gchar *dest = dest_slot->type;
 	img_to_slot_handler handler = NULL;
 
@@ -2658,21 +2789,13 @@ img_to_slot_handler get_update_handler(RaucImage *mfimage, RaucSlot *dest_slot,
 
 	g_message("Checking image type for slot type: %s", dest);
 
-	for (RaucUpdatePair *updatepair = updatepairs; updatepair->src != NULL; updatepair++) {
-		if (g_pattern_match_simple(updatepair->src, src) &&
-		    g_pattern_match_simple(updatepair->dest, dest)) {
-			g_message("Image detected as type: %s", updatepair->src);
-			handler = updatepair->handler;
-			break;
-		}
+	handler = get_handler_from_type(mfimage->type, dest);
+	if (!handler) {
+		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_NO_HANDLER,
+				"Unsupported image type '%s' for slot type '%s'", mfimage->type, dest);
+		return NULL;
 	}
 
-	if (handler == NULL) {
-		g_set_error(error, R_UPDATE_ERROR, R_UPDATE_ERROR_NO_HANDLER, "Unsupported image %s for slot type %s",
-				mfimage->filename, dest);
-		goto out;
-	}
-
-out:
+	g_message("Found handler for image type '%s' and slot type '%s'", mfimage->type, dest);
 	return handler;
 }
