@@ -146,6 +146,13 @@ have_streaming = pytest.mark.skipif(
 )
 
 
+def _have_qemu():
+    out, err, exitcode = run("grep -q qemu-test-init /proc/cmdline")
+    return exitcode == 0
+
+
+have_qemu = pytest.mark.skipif(not _have_qemu(), reason="Not in qemu-test")
+
 no_service = pytest.mark.skipif(string_in_config_h("ENABLE_SERVICE 1"), reason="Have service")
 
 
@@ -159,25 +166,29 @@ needs_emmc = pytest.mark.skipif("RAUC_TEST_EMMC" not in os.environ, reason="Miss
 needs_composefs = pytest.mark.skipif(not string_in_config_h("ENABLE_COMPOSEFS 1"), reason="Missing composefs support")
 
 
-def softhsm2_load_key_pair(cert, privkey, label, id_, softhsm2_mod):
+def softhsm2_load_key_pair(cert, privkey, label, id_, softhsm2_mod, tmp_path):
     proc = subprocess.run(
-        f"openssl x509 -in {cert} -inform pem -outform der "
-        f"| pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y cert -w /proc/self/fd/0 "
-        f"--label {label} --id {id_}",
+        f"pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y cert -w {cert} --label {label} --id {id_}",
         shell=True,
     )
     assert proc.returncode == 0
+
+    pubkey_file = tmp_path / f"pubkey_{label}.pem"
+
     proc = subprocess.run(
-        f"openssl rsa -in {privkey} -inform pem -pubout -outform der "
-        f"| pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y pubkey -w /proc/self/fd/0 "
-        f"--label {label} --id {id_}",
+        f"openssl rsa -in {privkey} -inform pem -pubout -outform pem -out {pubkey_file}",
         shell=True,
     )
     assert proc.returncode == 0
+
     proc = subprocess.run(
-        f"openssl rsa -in {privkey} -inform pem -outform der "
-        f"| pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y privkey -w /proc/self/fd/0 "
-        f"--label {label} --id {id_}",
+        f"pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y pubkey -w {pubkey_file} --label {label} --id {id_}",
+        shell=True,
+    )
+    assert proc.returncode == 0
+
+    proc = subprocess.run(
+        f"pkcs11-tool --module {softhsm2_mod} -l --pin 1111 -y privkey -w {privkey} --label {label} --id {id_}",
         shell=True,
     )
     assert proc.returncode == 0
@@ -226,6 +237,7 @@ def prepare_softhsm2(tmp_path, softhsm2_mod):
     ca_dev = Path("openssl-ca/dev")
     ca_cert = Path("openssl-ca/dev-ca.pem")
     enc_keys = Path("openssl-enc/keys")
+    web_ca = Path("openssl-ca/web")
 
     softhsm2_conf = tmp_path / "softhsm2.conf"
     softhsm2_dir = tmp_path / "softhsm2.tokens"
@@ -251,10 +263,20 @@ def prepare_softhsm2(tmp_path, softhsm2_mod):
 
     # load signing key pairs
     softhsm2_load_key_pair(
-        ca_dev / "autobuilder-1.cert.pem", ca_dev / "private/autobuilder-1.pem", "autobuilder-1", "01", softhsm2_mod
+        ca_dev / "autobuilder-1.cert.pem",
+        ca_dev / "private/autobuilder-1.pem",
+        "autobuilder-1",
+        "01",
+        softhsm2_mod,
+        tmp_path,
     )
     softhsm2_load_key_pair(
-        ca_dev / "autobuilder-2.cert.pem", ca_dev / "private/autobuilder-2.pem", "autobuilder-2", "02", softhsm2_mod
+        ca_dev / "autobuilder-2.cert.pem",
+        ca_dev / "private/autobuilder-2.pem",
+        "autobuilder-2",
+        "02",
+        softhsm2_mod,
+        tmp_path,
     )
 
     subprocess.check_call(
@@ -276,10 +298,21 @@ def prepare_softhsm2(tmp_path, softhsm2_mod):
         "enc-rsa-000",
         "11",
         softhsm2_mod,
+        tmp_path,
     )
 
     softhsm2_test_encryption(tmp_path, enc_keys / "rsa-4096/cert-000.pem", "enc-rsa-000")
 
+    # load web client key pair
+    softhsm2_load_key_pair(
+        web_ca / "client-1.cert.pem",
+        web_ca / "private/client-1.pem",
+        "client-1",
+        "21",
+        softhsm2_mod,
+        tmp_path,
+    )
+
     subprocess.check_call(f"pkcs11-tool --module {softhsm2_mod} -l --pin 1111 --list-objects", shell=True)
 
     os.environ["RAUC_PKCS11_PIN"] = "1111"
@@ -503,6 +536,7 @@ class System:
             "bootloader": "grub",
             "grubenv": "grubenv.test",
             "variant-name": "Default Variant",
+            "data-directory": "data-dir",
         }
         self.config["keyring"] = {
             "path": "openssl-ca/dev-ca.pem",
@@ -546,7 +580,6 @@ class System:
 
     def prepare_adaptive_config(self):
         self.prepare_minimal_config()
-        self.config["system"]["data-directory"] = "data-dir"
         self.config["handlers"] = {
             "system-info": "bin/systeminfo.sh",
             "pre-install": "bin/preinstall.sh",
@@ -635,17 +668,51 @@ def system(tmp_path, dbus_session_bus):
 
 
 class HTTPServer:
-    BASE = "http://127.0.0.1/backend"
-
     def __init__(self):
-        self.url = f"{self.BASE}/get"
+        self.server = None
+        # in the qemu test environment, the server is already running
+        if "RAUC_TEST_HTTP_BACKEND" in os.environ:
+            self.base = "http://127.0.0.1/backend"
+        else:
+            self.base = "http://127.0.0.1:8080"
+            self.start()
+        self.url = f"{self.base}/get"
+
+    def start(self):
+        if "RAUC_TEST_HTTP_BACKEND" in os.environ:
+            return
+        assert self.server is None
+
+        self.server = subprocess.Popen(["python3", "nginx_backend.py"])
+        timeout = time.monotonic() + 5.0
+        while True:
+            time.sleep(0.1)
+            try:
+                resp = requests.get(f"{self.base}/")
+                resp.raise_for_status()
+                break
+            except requests.exceptions.ConnectionError:
+                if time.monotonic() > timeout:
+                    raise
+
+    def stop(self):
+        if "RAUC_TEST_HTTP_SERVER" in os.environ:
+            return
+        assert self.server is not None
+
+        self.server.terminate()
+        try:
+            self.server.wait(timeout=10)
+        except subprocess.TimeoutExpired:
+            self.server.kill()
+            self.server.wait()
 
     def setup(self, *, file_path):
         resp = requests.post(
-            f"{self.BASE}/setup",
+            f"{self.base}/setup",
             timeout=5,
             json={
-                "file_path": file_path,
+                "file_path": os.path.abspath(file_path),
             },
         )
         resp.raise_for_status()
@@ -659,11 +726,13 @@ class HTTPServer:
         return requests.head(self.url, **kwargs)
 
     def get_summary(self):
-        resp = requests.get(f"{self.BASE}/summary", timeout=15)
+        resp = requests.get(f"{self.base}/summary", timeout=15)
         resp.raise_for_status()
         return resp.json()
 
 
-@pytest.fixture
-def http_server():
-    return HTTPServer()
+@pytest.fixture(scope="session")
+def http_server(env_setup):
+    server = HTTPServer()
+    yield server
+    server.stop()
