@@ -26,7 +26,7 @@ typedef enum {
 	TEST_UPDATE_HANDLER_INSTALL_HOOK                                  = BIT(6),
 	TEST_UPDATE_HANDLER_NO_HOOK_FILE                                  = BIT(7),
 	TEST_UPDATE_HANDLER_HOOK_FAIL                                     = BIT(8),
-	TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX                           = BIT(9),
+	TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX                       = BIT(9),
 	TEST_UPDATE_HANDLER_IMAGE_TOO_LARGE                               = BIT(10),
 } TestUpdateHandlerParams;
 
@@ -57,6 +57,7 @@ static void test_get_update_handler(UpdateHandlerFixture *fixture, gconstpointer
 	image = r_new_image();
 	image->slotclass = g_strdup("rootfs");
 	image->filename = g_strconcat("rootfs.", test_pair->imagetype, NULL);
+	image->type = g_strdup(derive_image_type_from_filename_pattern(image->filename));
 
 	targetslot = g_new0(RaucSlot, 1);
 	targetslot->name = g_intern_string("rootfs.0");
@@ -91,6 +92,7 @@ static void test_get_custom_update_handler(UpdateHandlerFixture *fixture, gconst
 	image->slotclass = g_strdup("rootfs");
 	image->filename = g_strdup("rootfs.custom");
 	image->hooks.install = TRUE;
+	image->type = g_strdup(derive_image_type_from_filename_pattern(image->filename));
 
 	targetslot = g_new0(RaucSlot, 1);
 	targetslot->name = g_intern_string("rootfs.0");
@@ -123,11 +125,11 @@ static void update_handler_fixture_set_up(UpdateHandlerFixture *fixture,
 	}
 
 	if ((test_pair->params & TEST_UPDATE_HANDLER_IMAGE_TOO_LARGE) &&
-	    (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX)) {
+	    (test_pair->params & TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX)) {
 		g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
 				"Checking image type for slot type: *");
 		g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
-				"Image detected as type: *");
+				"Found handler for image type * and slot type *");
 		g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_INFO,
 				"Selected adaptive update method *");
 		g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
@@ -147,7 +149,7 @@ static void update_handler_fixture_tear_down(UpdateHandlerFixture *fixture,
 	if (!(test_pair->params & TEST_UPDATE_HANDLER_NO_TARGET_DEV)) {
 		g_assert(test_remove(fixture->tmpdir, "rootfs-0") == 0);
 	}
-	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+	if (test_pair->params & TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX) {
 		test_rm_tree(fixture->tmpdir, "rootfs-0-datadir");
 	}
 	g_assert(test_rmdir(fixture->tmpdir, "") == 0);
@@ -395,7 +397,14 @@ static void test_update_handler(UpdateHandlerFixture *fixture,
 	/* create source image */
 	image = r_new_image();
 	image->slotclass = g_strdup("rootfs");
-	image->filename = g_strdup(imagepath);
+	if (g_strcmp0(test_pair->imagetype, "emptyfs") == 0) {
+		/* For emptyfs, don't set a filename since no file exists */
+		image->filename = NULL;
+		image->type = g_strdup("emptyfs");
+	} else {
+		image->filename = g_strdup(imagepath);
+		image->type = g_strdup(derive_image_type_from_filename_pattern(image->filename));
+	}
 	image->checksum.size = image_size;
 	image->checksum.digest = g_strdup("0xdeadbeef");
 	if (test_pair->params & TEST_UPDATE_HANDLER_HOOKS) {
@@ -413,7 +422,7 @@ static void test_update_handler(UpdateHandlerFixture *fixture,
 			test_do_chmod(tmp_filename);
 		}
 	}
-	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+	if (test_pair->params & TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX) {
 		image->adaptive = g_strsplit("block-hash-index", " ", 0);
 	}
 
@@ -441,7 +450,8 @@ static void test_update_handler(UpdateHandlerFixture *fixture,
 
 		r_context_conf()->install_info->mounted_bundle = g_new0(RaucBundle, 1);
 		r_context_conf()->install_info->mounted_bundle->storepath = storepath;
-	} else {
+	} else if (g_strcmp0(test_pair->imagetype, "emptyfs") != 0) {
+		/* No image file is needed emptyfs, so we just want to avoid dropping out here */
 		g_assert_not_reached();
 	}
 
@@ -453,7 +463,7 @@ no_image:
 	targetslot->device = g_strdup(slotpath);
 	targetslot->type = g_strdup(test_pair->slottype);
 	targetslot->state = ST_INACTIVE;
-	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+	if (test_pair->params & TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX) {
 		targetslot->data_directory = g_build_filename(fixture->tmpdir, "rootfs-0-datadir", NULL);
 	}
 
@@ -495,6 +505,44 @@ no_image:
 	} else if (g_strcmp0(test_pair->imagetype, "ext4") == 0) {
 		g_assert_cmpint(get_file_size(imagepath, NULL), ==, image_size);
 		g_assert(test_mount(slotpath, mountprefix));
+		g_assert(r_umount(slotpath, NULL));
+	} else if (g_strcmp0(test_pair->imagetype, "emptyfs") == 0) {
+		/* Verify emptyfs created an empty ext4 filesystem */
+		g_assert(test_mount(slotpath, mountprefix));
+
+		/* Check that only lost+found directory exists (standard for empty ext4) */
+		g_autoptr(GDir) dir = NULL;
+		GError *dir_error = NULL;
+		const gchar *entry;
+		gint entry_count = 0;
+		gboolean found_lost_found = FALSE;
+
+		dir = g_dir_open(mountprefix, 0, &dir_error);
+		g_assert_no_error(dir_error);
+		g_assert_nonnull(dir);
+
+		while ((entry = g_dir_read_name(dir)) != NULL) {
+			entry_count++;
+			if (g_strcmp0(entry, "lost+found") == 0) {
+				found_lost_found = TRUE;
+			}
+		}
+
+		/* Empty ext4 should only contain lost+found directory */
+		g_assert_cmpint(entry_count, ==, 1);
+		g_assert_true(found_lost_found);
+
+		/* Verify lost+found is empty */
+		g_autofree gchar *lost_found_path = g_build_filename(mountprefix, "lost+found", NULL);
+		g_autoptr(GDir) lost_found_dir = g_dir_open(lost_found_path, 0, &dir_error);
+		g_assert_no_error(dir_error);
+		g_assert_nonnull(lost_found_dir);
+		g_assert_null(g_dir_read_name(lost_found_dir));
+
+		/* Close directory handles before unmounting */
+		g_clear_pointer(&lost_found_dir, g_dir_close);
+		g_clear_pointer(&dir, g_dir_close);
+
 		g_assert(r_umount(slotpath, NULL));
 	} else if ((g_strcmp0(test_pair->imagetype, "tar") == 0) || ((g_strcmp0(test_pair->imagetype, "caidx") == 0))) {
 		g_autofree gchar *testpath = g_build_filename(mountprefix, "testfile.txt", NULL);
@@ -504,7 +552,7 @@ no_image:
 	}
 
 	/* check statistics */
-	if (test_pair->params & TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX) {
+	if (test_pair->params & TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX) {
 		RaucStats *stats;
 		guint64 count_zero = 0;
 		guint64 sum_zero = 0;
@@ -686,9 +734,9 @@ int main(int argc, char *argv[])
 		{"nor", "img", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
 
 		/* adaptive tests */
-		{"raw", "img", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
-		{"ext4", "img", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
-		{"raw", "ext4", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX, 0, 0},
+		{"raw", "img", TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX, 0, 0},
+		{"ext4", "img", TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX, 0, 0},
+		{"raw", "ext4", TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX, 0, 0},
 
 		/* casync blob index tests */
 		{"ext4", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
@@ -699,7 +747,23 @@ int main(int argc, char *argv[])
 
 		/* image too large */
 		{"ext4", "ext4", TEST_UPDATE_HANDLER_IMAGE_TOO_LARGE | TEST_UPDATE_HANDLER_EXPECT_FAIL, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED},
-		{"ext4", "ext4", TEST_UPDATE_HANDLER_INCR_BLOCK_HASH_IDX | TEST_UPDATE_HANDLER_IMAGE_TOO_LARGE | TEST_UPDATE_HANDLER_EXPECT_FAIL, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED},
+		{"ext4", "ext4", TEST_UPDATE_HANDLER_ADAPTIVE_BLOCK_HASH_IDX | TEST_UPDATE_HANDLER_IMAGE_TOO_LARGE | TEST_UPDATE_HANDLER_EXPECT_FAIL, R_UPDATE_ERROR, R_UPDATE_ERROR_FAILED},
+
+		{"nor", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"nand", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"vfat", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"jffs2", "img.caibx", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"jffs2", "img", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+
+		/* empty filesystem test (no image included) */
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_DEFAULT, 0, 0},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_NO_TARGET_DEV | TEST_UPDATE_HANDLER_EXPECT_FAIL, G_SPAWN_EXIT_ERROR, 1},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_PRE_HOOK, 0, 0},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_POST_HOOK, 0, 0},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_INSTALL_HOOK, 0, 0},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_PRE_HOOK | TEST_UPDATE_HANDLER_HOOK_FAIL | TEST_UPDATE_HANDLER_EXPECT_FAIL, G_SPAWN_EXIT_ERROR, 1},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_POST_HOOK | TEST_UPDATE_HANDLER_HOOK_FAIL | TEST_UPDATE_HANDLER_EXPECT_FAIL, G_SPAWN_EXIT_ERROR, 1},
+		{"ext4", "emptyfs", TEST_UPDATE_HANDLER_HOOKS | TEST_UPDATE_HANDLER_INSTALL_HOOK | TEST_UPDATE_HANDLER_HOOK_FAIL | TEST_UPDATE_HANDLER_EXPECT_FAIL, G_SPAWN_EXIT_ERROR, 1},
 
 		{0}
 	};
@@ -1133,5 +1197,85 @@ int main(int argc, char *argv[])
 			test_update_handler,
 			update_handler_fixture_tear_down);
 
+	g_test_add("/update_handler/get_handler/img.caibx_to_nor",
+			UpdateHandlerFixture,
+			&testpair_matrix[65],
+			NULL,
+			test_get_update_handler,
+			NULL);
+	g_test_add("/update_handler/get_handler/img.caibx_to_nand",
+			UpdateHandlerFixture,
+			&testpair_matrix[66],
+			NULL,
+			test_get_update_handler,
+			NULL);
+	g_test_add("/update_handler/get_handler/img.caibx_to_vfat",
+			UpdateHandlerFixture,
+			&testpair_matrix[67],
+			NULL,
+			test_get_update_handler,
+			NULL);
+	g_test_add("/update_handler/get_handler/img.caibx_to_jffs2",
+			UpdateHandlerFixture,
+			&testpair_matrix[68],
+			NULL,
+			test_get_update_handler,
+			NULL);
+	g_test_add("/update_handler/get_handler/img_to_jffs2",
+			UpdateHandlerFixture,
+			&testpair_matrix[69],
+			NULL,
+			test_get_update_handler,
+			NULL);
+	/* emptyfs tests */
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/default",
+			UpdateHandlerFixture,
+			&testpair_matrix[70],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/no-slot",
+			UpdateHandlerFixture,
+			&testpair_matrix[71],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/pre-hook",
+			UpdateHandlerFixture,
+			&testpair_matrix[72],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/post-hook",
+			UpdateHandlerFixture,
+			&testpair_matrix[73],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/install-hook",
+			UpdateHandlerFixture,
+			&testpair_matrix[74],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/pre-hook/fail",
+			UpdateHandlerFixture,
+			&testpair_matrix[75],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/post-hook/fail",
+			UpdateHandlerFixture,
+			&testpair_matrix[76],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+	g_test_add("/update_handler/update_handler/emptyfs_to_ext4/install-hook/fail",
+			UpdateHandlerFixture,
+			&testpair_matrix[77],
+			update_handler_fixture_set_up,
+			test_update_handler,
+			update_handler_fixture_tear_down);
+
 	return g_test_run();
 }
