@@ -317,7 +317,7 @@ static void environ_test(void)
 
 static void semver_parse_test(void)
 {
-	g_autofree gchar *version_string;
+	g_autofree gchar *version_string = NULL;
 	g_autofree guint64 version[3] = {0};
 	g_autofree gchar *pre_release = NULL;
 	g_autofree gchar *build = NULL;
@@ -506,6 +506,65 @@ static void format_duration_test(void)
 	g_free(tmp);
 }
 
+static void regex_match_test(void)
+{
+	gchar *tmp = NULL;
+
+	tmp = r_regex_match_simple("rauc\\.slot=(\\S+)", "root=foo quiet rauc.slot=system0 dummy");
+	g_assert_cmpstr(tmp, ==, "system0");
+	g_free(tmp);
+
+	tmp = r_regex_match_simple("rauc\\.slot=(\\S+)", " root=/dev/null ");
+	g_assert_null(tmp);
+}
+
+static void tempfile_cleanup_test(void)
+{
+	g_autofree gchar *tmpdir = g_dir_make_tmp("rauc-XXXXXX", NULL);
+
+	g_auto(RTempFile) filename = g_build_filename(tmpdir, "file", NULL);
+	g_assert_false(g_file_test(filename, G_FILE_TEST_EXISTS));
+
+	/* test missing file */
+	{
+		g_auto(RTempFile) temp_filename = g_strdup(filename);
+		(void)temp_filename;
+	}
+	g_assert_false(g_file_test(filename, G_FILE_TEST_EXISTS));
+
+	/* test normal file */
+	{
+		g_auto(RTempFile) temp_filename = g_strdup(filename);
+		g_auto(filedesc) temp_fd = g_open(temp_filename, O_RDWR|O_CLOEXEC|O_CREAT, 0);
+		(void)temp_fd;
+		g_assert_true(g_file_test(temp_filename, G_FILE_TEST_EXISTS));
+	}
+	g_assert_false(g_file_test(filename, G_FILE_TEST_EXISTS));
+
+	/* test directory instead of file (should warn) */
+	g_test_expect_message(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "failed to remove */file");
+	{
+		g_auto(RTempFile) temp_filename = g_strdup(filename);
+		g_assert_cmpint(g_mkdir(temp_filename, 0777), ==, 0);
+		g_assert_true(g_file_test(temp_filename, G_FILE_TEST_EXISTS));
+	}
+	g_assert_cmpint(g_rmdir(filename), ==, 0);
+	g_test_assert_expected_messages();
+}
+
+static void boottime_test(void)
+{
+	gint64 a = r_get_boottime();
+	g_usleep(100000);
+	gint64 b = r_get_boottime();
+
+	g_assert_cmpint(b, >, a);
+	gint64 diff = b - a;
+
+	g_assert_cmpint(diff, >=, 100000);
+	g_assert_cmpint(diff, <=, 110000);
+}
+
 int main(int argc, char *argv[])
 {
 	setlocale(LC_ALL, "C");
@@ -524,6 +583,9 @@ int main(int argc, char *argv[])
 	g_test_add_func("/utils/semver_parse_test", semver_parse_test);
 	g_test_add_func("/utils/semver_less_equal_test", semver_less_equal_test);
 	g_test_add_func("/utils/format_duration", format_duration_test);
+	g_test_add_func("/utils/regex_match", regex_match_test);
+	g_test_add_func("/utils/tempfile_cleanup", tempfile_cleanup_test);
+	g_test_add_func("/utils/boottime", boottime_test);
 
 	return g_test_run();
 }
